# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from typing import Optional

from dbnd._core.log import dbnd_log


if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.settings import DatabandSettings
    from dbnd._core.task import Task
    from dbnd._core.task_run.task_run import TaskRun


def get_databand_context():
    # type: () -> DatabandContext
    """Databand Context get or create instance."""
    from dbnd._core.context.databand_context import DatabandContext as _DatabandContext

    return _DatabandContext.try_instance()


def try_get_databand_context():
    # type: () -> Optional[DatabandContext]

    from dbnd._core.context.databand_context import DatabandContext as _DatabandContext

    if not _DatabandContext.has_instance():
        return None
    return get_databand_context()


def dbnd_context():
    """Get or create databand context."""
    context = try_get_databand_context()
    if not context:
        # we are running without Databand Context
        # let create one inplace
        from dbnd._core.context.databand_context import (
            DatabandContext as _DatabandContext,
        )

        context = _DatabandContext.try_instance(name="inplace_run")
    return context


def get_databand_run():
    # type: () -> DatabandRun
    """Returns current Task/Pipeline/Flow instance."""
    from dbnd._core.run.databand_run import DatabandRun as _DatabandRun

    v = _DatabandRun.get_instance()
    return v


def try_get_databand_run():
    # type: () -> Optional[DatabandRun]
    from dbnd._core.run.databand_run import DatabandRun as _DatabandRun

    if _DatabandRun.has_instance():
        return get_databand_run()
    return None


def in_tracking_run():
    # type: () -> bool
    run = try_get_databand_run()
    if run:
        return not run.is_orchestration

    return False


def is_orchestration_run():
    # type: () -> bool
    run = try_get_databand_run()
    if run:
        return run.is_orchestration

    return False


def current_task():
    # type: () -> Task
    """
    Returns the current task's object.

    Example::

        from dbnd import current_task, task

        @task
        def calculate_alpha(alpha: int = 0.5):
            return current_task().task_version
    """
    from dbnd._core.task_build.task_context import current_task as ct

    return ct()


def try_get_current_task():
    from dbnd._core.task_build.task_context import try_get_current_task as tgct

    return tgct()


def get_task_by_task_id(task_id):
    # type: (str) -> Task
    return get_databand_context().task_instance_cache.get_task_by_id(task_id)


def try_get_current_task_run():
    # type: () -> TaskRun
    run = try_get_databand_run()
    if not run:
        return None
    task = try_get_current_task()
    if task:
        return run.get_task_run(task.task_id)
    else:
        return None


def current_task_run():
    # type: () -> TaskRun
    """
    Returns the current task run.

    Example::

        from dbnd import current_task_run, task

        @task
        def calculate_alpha(alpha: int = 0.5):
            return current_task_run()
    """
    return get_databand_run().get_task_run(current_task().task_id)


def get_settings():
    # type: () -> DatabandSettings
    v = get_databand_context().settings  # type: DatabandSettings
    return v


def is_verbose():
    return dbnd_log.is_verbose()


def get_target_logging_level():
    default_level = 10  # DEBUG

    context = try_get_databand_context()
    if context and getattr(context, "settings", None):
        return getattr(logging, context.settings.log.targets_log_level)

    return default_level


def is_killed():
    run = try_get_databand_run()
    return run and run.is_killed()


def cancel_current_run(message=None):
    """Kills a run's execution from within the execution."""
    current_databand_run = get_databand_run()
    return current_databand_run.kill_run(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import sys
import re
import collections
import copy


__version__ = '1.5.1'


class Pyasciigraph:

    def __init__(self, line_length=79,
                 min_graph_length=50,
                 separator_length=2,
                 force_max_value=None,
                 graphsymbol=None,
                 multivalue=True,
                 human_readable=None,
                 float_format='{0:.0f}',
                 titlebar='#'
                 ):
        """Constructor of Pyasciigraph

        :param line_length: the max number of char on a line
          if any line cannot be shorter,
          it will go over this limit.
          Default: 79
        :type line_length: int
        :param min_graph_length: the min number of char
          used by the graph itself.
          Default: 50
        :type min_graph_length: int
        :param force_max_value: if provided, force a max value in order to graph
          each line with respect to it (only taking the actual max value if
          it is greater).
        :type: force_max_value: int
        :param separator_length: the length of field separator.
          Default: 2
        :type separator_length: int
        :param graphsymbol: the symbol used for the graph bar.
          Default: '█'
        :type graphsymbol: str or unicode (length one)
        :param multivalue: displays all the values if multivalued when True.
          displays only the max value if False
          Default: True
        :type multivalue: boolean
        :param human_readable: trigger human readable display (K, G, etc)
          Default: None (raw value display)

          * 'si' for power of 1000

          * 'cs' for power of 1024

          * any other value for raw value display)

        :type human_readable: string (si, cs, none)
        :param float_format: formatting of the float value
          Default: '{0:.0f}' (convert to integers).
          expample: '{:,.2f}' (2 decimals, '.' to separate decimal and int,
          ',' every three power of tens).
        :param titlebar: sets the character(s) for the horizontal title bar
          Default: '#'
        :type titlebar: string
        """

        self.line_length = line_length
        self.separator_length = separator_length
        self.min_graph_length = min_graph_length
        self.max_value = force_max_value
        self.float_format = float_format
        self.titlebar = titlebar
        if graphsymbol is None:
            self.graphsymbol = self._u('█')
        else:
            self.graphsymbol = graphsymbol
        if self._len_noansi(self.graphsymbol) != 1:
            raise Exception('Bad graphsymbol length, must be 1',
                            self._len_noansi(self.graphsymbol))
        self.multivalue = multivalue
        self.hsymbols = [self._u(''), self._u('K'), self._u('M'),
                         self._u('G'), self._u('T'), self._u('P'),
                         self._u('E'), self._u('Z'), self._u('Y')]

        if human_readable == 'si':
            self.divider = 1000
        elif human_readable == 'cs':
            self.divider = 1024
        else:
            self.divider = None

    @staticmethod
    def _len_noansi(string):
        l = len(re.sub('\x1b[^m]*m', '', string))
        return l

    def _trans_hr(self, value):

        if self.divider is None:
            return self.float_format.format(value)
        vl = value
        for hs in self.hsymbols:
            new_val = vl / self.divider
            if new_val < 1:
                return self.float_format.format(vl) + hs
            else:
                vl = new_val
        return self.float_format.format(vl * self.divider) + hs

    @staticmethod
    def _u(x):
        """Unicode compat helper
        """
        if sys.version < '3':
            return x + ''.decode("utf-8")
        else:
            return x

    @staticmethod
    def _color_string(string, color):
        """append color to a string + reset to white at the end of the string
        """
        if color is None:
            return string
        else:
            return color + string + '\033[0m'

    def _get_thresholds(self, data):
        """get various info (min, max, width... etc)
        from the data to graph.
        """
        all_thre = {}
        all_thre['value_max_length'] = 0
        all_thre['info_max_length'] = 0
        all_thre['max_pos_value'] = 0
        all_thre['min_neg_value'] = 0

        if self.max_value is not None:
            all_thre['max_pos_value'] = self.max_value

        # Iterate on all the items
        for (info, value, color) in data:
            totalvalue_len = 0

            # If we have a list of values for the item
            if isinstance(value, collections.Iterable):
                icount = 0
                maxvalue = 0
                minvalue = 0
                for (ivalue, icolor) in value:
                    if ivalue < minvalue:
                        minvalue = ivalue
                    if ivalue > maxvalue:
                        maxvalue = ivalue
                    # if we are in multivalued mode, the value string is
                    # the concatenation of the values, separeted by a ',',
                    # len() must be computed on it
                    # if we are not in multivalued mode, len() is just the
                    # longer str(value) len ( /!\, value can be negative,
                    # which means that it's not simply len(str(max_value)))
                    if self.multivalue:
                        totalvalue_len += len("," + self._trans_hr(ivalue))
                    else:
                        totalvalue_len = max(totalvalue_len, len(self._trans_hr(ivalue)))

                if self.multivalue:
                    # remove one comma if multivalues
                    totalvalue_len = totalvalue_len - 1

            # If the item only has one value
            else:
                totalvalue_len = len(self._trans_hr(value))
                maxvalue = value
                minvalue = value

            if minvalue < all_thre['min_neg_value']:
                all_thre['min_neg_value'] = minvalue

            if maxvalue > all_thre['max_pos_value']:
                all_thre['max_pos_value'] = maxvalue

            if self._len_noansi(info) > all_thre['info_max_length']:
                all_thre['info_max_length'] = self._len_noansi(info)

            if totalvalue_len > all_thre['value_max_length']:
                all_thre['value_max_length'] = totalvalue_len

        return all_thre

    def _gen_graph_string(
            self, value, max_value, min_neg_value, graph_length, start_value_pos, color):
        """Generate the bar + its paddings (left and right)
        """
        def _gen_graph_string_part(
                value, max_value, min_neg_value, graph_length, color):

            all_width = max_value + abs(min_neg_value)

            if all_width == 0:
                bar_width = 0
            else:
                bar_width = int(abs(float(value)) * float(graph_length) / float(all_width))

            return (Pyasciigraph._color_string(
                    self.graphsymbol * bar_width,
                color),
                bar_width
                )

        all_width = max_value + abs(min_neg_value)

        if all_width == 0:
            bar_width = 0
            neg_width = 0
            pos_width = 0
        else:
            neg_width = int(abs(float(min_neg_value)) * float(graph_length) / float(all_width))
            pos_width = int(abs(max_value) * graph_length / all_width)

        if isinstance(value, collections.Iterable):
            accuvalue = 0
            totalstring = ""
            totalsquares = 0

            sortedvalue = copy.deepcopy(value)
            sortedvalue.sort(reverse=False, key=lambda tup: tup[0])
            pos_value = [x for x in sortedvalue if x[0] >= 0]
            neg_value = [x for x in sortedvalue if x[0] < 0]

            # for the negative values, we build the bar + padding from 0 to the left
            for i in reversed(neg_value):
                ivalue = i[0]
                icolor = i[1]
                scaled_value = ivalue - accuvalue
                (partstr, squares) = _gen_graph_string_part(
                    scaled_value, max_value, min_neg_value, graph_length, icolor)
                totalstring = partstr + totalstring
                totalsquares += squares
                accuvalue += scaled_value

            # left padding
            totalstring = Pyasciigraph._u(' ') * (neg_width - abs(totalsquares)) + totalstring

            # reset some counters
            accuvalue = 0
            totalsquares = 0

            # for the positive values we build the bar from 0 to the right
            for i in pos_value:
                ivalue = i[0]
                icolor = i[1]
                scaled_value = ivalue - accuvalue
                (partstr, squares) = _gen_graph_string_part(
                    scaled_value, max_value, min_neg_value, graph_length, icolor)
                totalstring += partstr
                totalsquares += squares
                accuvalue += scaled_value

            # right padding
            totalstring += Pyasciigraph._u(' ') * (start_value_pos - neg_width - abs(totalsquares))
            return totalstring
        else:
            # handling for single value item
            (partstr, squares) = _gen_graph_string_part(
                value, max_value, min_neg_value, graph_length, color)
            if value >= 0:
                return Pyasciigraph._u(' ') * neg_width + \
                        partstr + \
                        Pyasciigraph._u(' ') * (start_value_pos - (neg_width + squares))
            else:
                return Pyasciigraph._u(' ') * (neg_width - squares) +\
                        partstr +\
                        Pyasciigraph._u(' ') * (start_value_pos - neg_width)


    def _gen_info_string(self, info, start_info_pos, line_length):
        """Generate the info string + padding
        """
        number_of_space = (line_length - start_info_pos - self._len_noansi(info))
        return info + Pyasciigraph._u(' ') * number_of_space

    def _gen_value_string(self, value, min_neg_value, color, start_value_pos, start_info_pos):
        """Generate the value string + padding
        """
        icount = 0
        if isinstance(value, collections.Iterable) and self.multivalue:
            for (ivalue, icolor) in value:
                if icount == 0:
                    # total_len is needed because the color characters count
                    # with the len() function even when they are not printed to
                    # the screen.
                    totalvalue_len = len(self._trans_hr(ivalue))
                    totalvalue = Pyasciigraph._color_string(
                        self._trans_hr(ivalue), icolor)
                else:
                    totalvalue_len += len("," + self._trans_hr(ivalue))
                    totalvalue += "," + \
                        Pyasciigraph._color_string(
                            self._trans_hr(ivalue),
                            icolor)
                icount += 1
        elif isinstance(value, collections.Iterable):
            max_value = min_neg_value
            color = None
            for (ivalue, icolor) in value:
                if ivalue > max_value:
                    max_value = ivalue
                    color = icolor
            totalvalue_len = len(self._trans_hr(max_value))
            totalvalue = Pyasciigraph._color_string(
                self._trans_hr(max_value), color)

        else:
            totalvalue_len = len(self._trans_hr(value))
            totalvalue = Pyasciigraph._color_string(
                self._trans_hr(value), color)

        number_space = start_info_pos -\
            start_value_pos -\
            totalvalue_len -\
            self.separator_length

        # This must not be negitive, this happens when the string length is
        # larger than the separator length
        if number_space < 0:
            number_space = 0

        return  ' ' * number_space + totalvalue +\
                ' ' * \
            ((start_info_pos - start_value_pos - totalvalue_len)
             - number_space)

    def _sanitize_string(self, string):
        """try to convert strings to UTF-8
        """
        # get the type of a unicode string
        unicode_type = type(Pyasciigraph._u('t'))
        input_type = type(string)
        if input_type is str:
            if sys.version < '3':
                info = unicode(string)
            else:
                info = string
        elif input_type is unicode_type:
            info = string
        elif input_type is int or input_type is float:
            if sys.version < '3':
                info = unicode(string)
            else:
                info = str(string)
        else:
            info = str(string)
        return info

    def _sanitize_value(self, value):
        """try to values to UTF-8
        """
        if isinstance(value, collections.Iterable):
            newcollection = []
            for i in value:
                if len(i) == 1:
                    newcollection.append((i[0], None))
                elif len(i) >= 2:
                    newcollection.append((i[0], i[1]))
            return newcollection
        else:
            return value

    def _sanitize_data(self, data):
        ret = []
        for item in data:
            if (len(item) == 2):
                if isinstance(item[1], collections.Iterable):
                    ret.append(
                        (self._sanitize_string(item[0]),
                         self._sanitize_value(item[1]),
                         None))
                else:
                    ret.append(
                        (self._sanitize_string(item[0]),
                         self._sanitize_value(item[1]),
                         None))
            if (len(item) == 3):
                ret.append(
                    (self._sanitize_string(item[0]),
                     self._sanitize_value(item[1]),
                     item[2]))
        return ret

    def graph(self, label=None, data=[]):
        """function generating the graph

        :param string label: the label of the graph
        :param iterable data: the data (list of tuple (info, value))
                info must be "castable" to a unicode string
                value must be an int or a float
        :rtype: a list of strings (each lines of the graph)

        """
        result = []
        san_data = self._sanitize_data(data)
        all_thre = self._get_thresholds(san_data)

        if not label is None:
            san_label = self._sanitize_string(label)
            label_len = self._len_noansi(san_label)
        else:
            label_len = 0

        real_line_length = max(self.line_length, label_len)

        min_line_length = self.min_graph_length +\
            2 * self.separator_length +\
            all_thre['value_max_length'] +\
            all_thre['info_max_length']

        if min_line_length < real_line_length:
            # calcul of where to start info
            start_info_pos = self.line_length -\
                all_thre['info_max_length']
            # calcul of where to start value
            start_value_pos = start_info_pos -\
                self.separator_length -\
                all_thre['value_max_length']
            # calcul of where to end graph
            graph_length = start_value_pos -\
                self.separator_length
        else:
            # calcul of where to start value
            start_value_pos = self.min_graph_length +\
                self.separator_length
            # calcul of where to start info
            start_info_pos = start_value_pos +\
                all_thre['value_max_length'] +\
                self.separator_length
            # calcul of where to end graph
            graph_length = start_value_pos -\
                self.separator_length
            # calcul of the real line length
            real_line_length = min_line_length

        if not label is None:
            result.append(san_label)
            result.append(Pyasciigraph._u(self.titlebar) * real_line_length)

        header_line = (Pyasciigraph._u(" ") * graph_length) + "count"
        header_line += Pyasciigraph._u(" ") * (start_info_pos - len(header_line)) + "bin"

        result.append(header_line)

        for info, value, color in san_data:

            graph_string = self._gen_graph_string(
                value,
                    all_thre['max_pos_value'],
                    all_thre['min_neg_value'],
                    graph_length,
                    start_value_pos,
                    color
            )

            value_string = self._gen_value_string(
                value,
                    all_thre['min_neg_value'],
                    color,
                    start_value_pos,
                    start_info_pos,
            )

            info_string = self._gen_info_string(
                info,
                    start_info_pos,
                    real_line_length
            )
            new_line = graph_string + value_string + info_string
            result.append(new_line)

        return result



# © Copyright Databand.ai, an IBM Company 2022

import abc
import logging
import typing

import six

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.constants import TaskEssence
from dbnd._core.plugin.dbnd_airflow_operator_plugin import (
    build_task_at_airflow_dag_context,
    is_in_airflow_dag_build_context,
)
from dbnd._core.task_build.task_context import try_get_current_task
from dbnd._core.task_build.task_definition import TaskDefinition
from dbnd._core.task_build.task_factory import TaskFactory
from dbnd._core.task_build.task_registry import get_task_registry


if typing.TYPE_CHECKING:
    from dbnd._core.task.base_task import _BaseTask

logger = logging.getLogger(__name__)


class TaskMetaclass(abc.ABCMeta):
    """
    The Metaclass of :py:class:`Task`.
    """

    def __new__(mcs, classname, bases, classdict):
        """
        Custom class creation for namespacing.

        Also register all subclasses.

        When the set or inherited namespace evaluates to ``None``, set the task namespace to
        whatever the currently declared namespace is.
        """
        cls = super(TaskMetaclass, mcs).__new__(
            mcs, classname, bases, classdict
        )  # type: typing.Type[_BaseTask]

        # we are starting from "not clean" classdict ->
        # A. it's deserialization
        # B. it was calculated before
        if classdict.get("task_definition") is not None:
            return cls

        cls.task_definition = TaskDefinition.from_task_cls(
            task_class=cls, classdict=classdict
        )

        # now we will assign all calculated parameters
        # so instead of ParameterFactory, we will have ParameterDefinition
        for k, v in six.iteritems(cls.task_definition.task_param_defs):
            setattr(cls, k, v)

        # every time we see new implementation, we want it to have an priority over old implementation
        # we need to switch to dict() and store history else where
        r = get_task_registry()
        r.register_task(cls)

        return cls

    def _build_task_obj(cls, **kwargs):
        # called from TaskFactory to create object
        return super(TaskMetaclass, cls).__call__(**kwargs)

    def __call__(cls, *args, **kwargs):
        """
        Custom class instantiation utilizing instance cache.
        """

        # use-case of TaskClass() call from airflow context during DAG creation
        _dbnd_disable_airflow_inplace = kwargs.pop(
            "_dbnd_disable_airflow_inplace", False
        )
        if (
            is_in_airflow_dag_build_context()
            and TaskEssence.is_task_cls(cls)
            and not _dbnd_disable_airflow_inplace
            and not getattr(cls, "_dbnd_decorated_task", False)
        ):
            kwargs = kwargs.copy()
            kwargs["_dbnd_disable_airflow_inplace"] = True
            return build_task_at_airflow_dag_context(
                task_cls=cls, call_args=args, call_kwargs=kwargs
            )

        task_definition = cls.task_definition
        # we need to have context initialized before we start to run all logic in config() scope

        # create new config layer, so when we are out of this process -> config is back to the previous value
        with config(
            config_values={},
            source=task_definition.task_passport.format_source_name("ctor"),
        ) as task_config:
            factory = TaskFactory(
                config=task_config,
                task_cls=cls,
                task_definition=cls.task_definition,
                task_args=args,
                task_kwargs=kwargs,
            )
            task_object = factory.build_task_object(cls)

        parent_task = try_get_current_task()
        if (
            parent_task
            and hasattr(task_object, "task_id")
            and (task_object.task_essence != TaskEssence.CONFIG)
        ):
            parent_task.descendants.add_child(task_object.task_id)

        return task_object



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys

import dbnd

from dbnd import task
from dbnd.tasks.doctor.doctor_report_builder import DoctorStatusReportBuilder


logger = logging.getLogger(__name__)


@task
def python_status(python_packages=False):
    """
    Shows the status of the python system
    :param python_packages: include explicit python packages info (pip freeze)
    :return:
    """
    logger.info("Running python checks..")
    report = DoctorStatusReportBuilder("Python Status")
    report.log("dbnd.version", dbnd.__version__)
    report.log("python.version", sys.version)
    report.log("python.path", sys.path)

    try:
        import pip

        report.log("pip.version", pip.__version__)
        try:
            from pip._internal.operations import freeze
        except ImportError:  # pip < 10.0
            from pip.operations import freeze
        packages = list(freeze.freeze())
        report.log("python.total_packages", len(packages))
        report.log("dbnd.packages", [p for p in packages if "dbnd" in p])

        # we want to show all packages only if explicitly requested
        if python_packages:
            logger.info("pip freeze: %s", packages)
            packages_page_size = 20  # we want to show only X packages per metric
            for i in range(0, len(packages), packages_page_size):
                report.log(
                    "python.packages.%s" % i, packages[i : i + packages_page_size]
                )
    except Exception as ex:
        logger.exception("Failed to get packages information via pips")

        report.log("python.packages", str(ex))
        report.log("python.total_packages", 0)
    return report.get_status_str()



# isort: skip_file
from datetime import datetime
from typing import Dict, Optional, Union

from dbnd._core.parameter.parameter_builder import ParameterFactory

def task(
    task_class_version: Optional[str] = None,
    task_env: Optional[str] = None,
    task_target_date: Optional[datetime.date] = None,
    task_airflow_op_kwargs: Optional[Dict[str, object]] = None,
    task_config: Optional[dict] = None,
    result: Optional = None,
    **kwargs,
):
    """
    :param task_class_version: Used to indicate persistent changes in a code.
    :param task_env: defines data and compute environment for a task. By default, tasks run in local environment. Part of task signature
    :param task_target_date: A date (data) associated with task execution. Part of task signature.
    Example value is 'datetime.date.today()'.
    :param task_airflow_op_kwargs:  Parameters to pass to Airflow operator that would run this task.
    :param task_config: A dictionary of arbitrary configuration params. How to override specific task sub configs
     task_config = {  "spark" : { "param" : "ss"  }
     task_config = { spark.jars = some_jars ,
                     kubernetes.gpu = some_gpu }
    :param result: Defines the name of the output of the task. When output is complex, like a tuple, we can name
    every item inside it. result=("output1", "output2")
    """
    ...

def pipeline(
    task_class_version: Optional[str] = None,
    task_env: Optional[str] = None,
    task_target_date: Optional[datetime.date] = None,
    task_airflow_op_kwargs: Optional[Dict[str, object]] = None,
    task_config: Optional[dict] = None,
    result: Optional = None,
    **kwargs,
):
    """
    :param task_class_version: Used to indicate persistent changes in a code.
    :param task_env: defines data and compute environment for a task. By default, tasks run in local environment. Part of task signature
    :param task_target_date: A date (data) associated with task execution. Part of task signature.
    Example value is 'datetime.date.today()'.
    :param task_airflow_op_kwargs:  Parameters to pass to Airflow operator that would run this task.
    :param task_config: A dictionary of arbitrary configuration params. How to override specific task sub configs
     task_config = {  "spark" : { "param" : "ss"  }
     task_config = { spark.jars = some_jars ,
                     kubernetes.gpu = some_gpu }
    :param result: Defines the name of the output of the task. When output is complex, like a tuple, we can name
    every item inside it. result=("output1", "output2")
    """
    ...

def band(
    task_class_version: Optional[str] = None,
    task_env: Optional[str] = None,
    task_target_date: Optional[datetime.date] = None,
    task_airflow_op_kwargs: Optional[Dict[str, object]] = None,
    task_config: Optional[dict] = None,
    result: Optional = None,
    **kwargs,
):
    """
    :param task_class_version: Used to indicate persistent changes in a code.
    :param task_env: defines data and compute environment for a task. By default, tasks run in local environment. Part of task signature
    :param task_target_date: A date (data) associated with task execution. Part of task signature.
    Example value is 'datetime.date.today()'.
    :param task_airflow_op_kwargs:  Parameters to pass to Airflow operator that would run this task.
    :param task_config: A dictionary of arbitrary configuration params. How to override specific task sub configs
     task_config = {  "spark" : { "param" : "ss"  }
     task_config = { spark.jars = some_jars ,
                     kubernetes.gpu = some_gpu }
    :param result: Defines the name of the output of the task. When output is complex, like a tuple, we can name
    every item inside it. result=("output1", "output2")
    """
    ...

def data_source_pipeline(
    task_env: Optional[str] = None,
    task_target_date: Optional[datetime.date] = None,
    task_config: Optional[dict] = None,
    **kwargs,
):
    """
    :param task_env: defines data and compute environment for a task. By default, tasks run in local environment. Part of task signature
    :param task_target_date: A date (data) associated with task execution. Part of task signature.
    Example value is 'datetime.date.today()'.
    :param task_config: A dictionary of arbitrary configuration params. How to override specific task sub configs
     task_config = {  "spark" : { "param" : "ss"  }
     task_config = { spark.jars = some_jars ,
                     kubernetes.gpu = some_gpu }
    """
    ...

_default_output: Union[ParameterFactory, object]

class TaskDecorator(object): ...

def build_task_decorator(*decorator_args, **decorator_kwargs): ...



from __future__ import absolute_import
"""
Copyright 2018 Steven Loria

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from distutils.version import LooseVersion


try:
    import marshmallow

    version_tuple = tuple(LooseVersion(marshmallow.__version__).version)
    if version_tuple != (2, 18, 0):
        raise ImportError

    from marshmallow import (
        base,
        class_registry,
        decorators,
        exceptions,
        fields,
        marshalling,
        orderedset,
        schema,
        utils,
        validate,
        warnings,
        warnings,
    )
    from marshmallow.schema import Schema, SchemaOpts, MarshalResult, UnmarshalResult
    from marshmallow.decorators import (
        pre_dump,
        post_dump,
        pre_load,
        post_load,
        validates,
        validates_schema,
    )
    from marshmallow.utils import pprint, missing
    from marshmallow.exceptions import ValidationError

except ImportError:
    from dbnd._vendor import _marshmallow as marshmallow

    from dbnd._vendor._marshmallow import (
        base,
        class_registry,
        decorators,
        exceptions,
        fields,
        marshalling,
        orderedset,
        schema,
        utils,
        validate,
        warnings,
        warnings,
    )
    from dbnd._vendor._marshmallow.schema import (
        Schema,
        SchemaOpts,
        MarshalResult,
        UnmarshalResult,
    )
    from dbnd._vendor._marshmallow.decorators import (
        pre_dump,
        post_dump,
        pre_load,
        post_load,
        validates,
        validates_schema,
    )
    from dbnd._vendor._marshmallow.utils import pprint, missing
    from dbnd._vendor._marshmallow.exceptions import ValidationError

__version__ = marshmallow.__version__
__version_info__ = marshmallow.__version_info__
__author__ = marshmallow.__author__

__all__ = [
    "Schema",
    "SchemaOpts",
    "fields",
    "validates",
    "validates_schema",
    "pre_dump",
    "post_dump",
    "pre_load",
    "post_load",
    "pprint",
    "MarshalResult",
    "UnmarshalResult",
    "ValidationError",
    "missing",
]



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.tracking.schemas.base import ApiStrictSchema
from dbnd._vendor.marshmallow import fields


class LogMessageSchema(ApiStrictSchema):
    source = fields.String(allow_none=True)
    stack_trace = fields.String(allow_none=True)
    timestamp = fields.DateTime(allow_none=True)
    dbnd_version = fields.String(allow_none=True, missing=None)



# © Copyright Databand.ai, an IBM Company 2022

import typing

from typing import Any, Set

import dbnd

from dbnd._core.errors import DatabandConfigError, ParseParameterError


if typing.TYPE_CHECKING:
    from dbnd._core.parameter.parameter_definition import ParameterDefinition


def wrong_func_attr_format(attr_path):
    return DatabandConfigError(
        "Wrong format for code path:  got '%s' instead of  module_ref.module.name"
        % attr_path
    )


def task_name_and_from_are_the_same(section, param):
    return DatabandConfigError(
        "Your [%s] defined using _type=%s. "
        "System can't handle recursive type resolving. "
        "Probably you are missing the module that has '%s' defined. "
        % (param, section, section),
        help_msg="Please check your configuration section [%s] "
        "or try to install package with %s" % (section, section),
    )


def env_section_not_found(section):
    return DatabandConfigError(
        "%s is listed in your `environments` list but a config section with this name could not be found"
        % section,
        help_msg="Please add a config section [%s] that configures the environment or remove it from the environment list"
        % section,
    )


def dbnd_root_local_not_defined(env):
    return DatabandConfigError(
        "You environment '{env}' definition doesn't have dbnd_local_root defined".format(
            env=env
        ),
        help_msg="Please define it at [{env}]. This path is used for all local internal dbdn system outputs".format(
            env=env
        ),
    )


def no_credentials():
    return DatabandConfigError(
        "No credentials are defined for the cloud environment you are trying to use",
        help_msg="Please configure the connection for your cloud env's conn_id (if conn_id is unset it will be [aws|gcp|azure]_default).\n"
        + "See: https://databand.readme.io/docs/integrations",
    )


def scheduled_job_missing_param(field):
    return DatabandConfigError("Scheduled job %s must be defined" % field)


def scheduled_job_exists(name):
    return DatabandConfigError(
        "Scheduled job named '%s' already exists" % name,
        help_msg="If you wish to update it add --update to the cli command",
    )


def scheduled_job_not_exists(name):
    return DatabandConfigError(
        "Scheduled job named '%s' does not exists and as such cannot be updated" % name
    )


def scheduled_job_invalid_interval(schedule_interval):
    from dbnd.api.shared_schemas.scheduled_job_schema import SCHEDULE_INTERVAL_PRESETS

    return DatabandConfigError(
        "The schedule_interval '%s' is not valid" % schedule_interval,
        help_msg="Valid values are any valid cron expression or one of the presets: %s"
        % ", ".join(SCHEDULE_INTERVAL_PRESETS),
    )


def wrong_store_name(name):
    return DatabandConfigError(
        "Unsupported tracking store: '{}', use one of file/console/api".format(name),
        help_msg="Please check you configuration at [core] tracker.",
    )


def wrong_tracking_api_name(name):
    return DatabandConfigError(
        "Unsupported tracking api: '{}', use one of web/db".format(name),
        help_msg="Please check you configuration at [core] tracker_api.",
    )


def choice_error(param, value, choices):
    # type: (ParameterDefinition,Any, Set[Any])->ParseParameterError
    return ParseParameterError(
        "{param}: {value} is not a valid choice from {choices}".format(
            param=param, value=value, choices="/".join(map(str, choices))
        )
    )


def empty_string_validator(param):
    # type: (ParameterDefinition)->ParseParameterError
    return ParseParameterError(
        "{param} has empty string value".format(param=param),
        help_msg="Please check your configuration for {}".format(param),
    )


def missing_module(module, reason=None):
    return DatabandConfigError(
        "'{module}' module is not found. {reason}".format(
            module=module, reason=reason or ""
        ),
        help_msg="Please, `pip install '{module}=={version}'`,".format(
            module=module, version=dbnd.__version__
        ),
    )



# © Copyright Databand.ai, an IBM Company 2022


class DataTargetCtrl(object):
    def __init__(self, target):
        """
        :param target: DataTarget
        """
        super(DataTargetCtrl, self).__init__()
        from targets.data_target import DataTarget

        self.target = target  # type: DataTarget



# © Copyright Databand.ai, an IBM Company 2022

import inspect
import itertools
import logging
import re
import typing

from typing import List, Optional

from dbnd._vendor.pytypes.type_utils import get_Union_params, is_Union
from targets.types import EagerLoad, LazyLoad
from targets.values.builtins_values import DefaultObjectValueType
from targets.values.structure import _StructureValueType
from targets.values.value_type import InlineValueType, ValueType


def f_name(f):
    return f.__name__


logger = logging.getLogger(__name__)


class ValueTypeRegistry(object):
    def __init__(self, known_value_types):
        self.value_types = []  # type: List[ValueType]
        self.discoverable_value_types = []  # type: List[ValueType]
        self.default = DefaultObjectValueType()

        # now for every parameter we also have text representation of the type
        # will be used for annotations
        self.type_str_to_parameter = {}
        for value_type in known_value_types:
            self.register_value_type(value_type)

        self._type_handler_from_type = TypeHandlerFromType(self)
        self._type_handler_from_type_str = TypeHandlerFromDocAnnotation(self)

    def register_value_type(self, value_type):
        self.value_types.append(value_type)
        if value_type.discoverable:
            self.discoverable_value_types.append(value_type)
        try:
            # for t in [value_type.type]:

            type_str = value_type.type_str
            if type_str:
                self.type_str_to_parameter[type_str] = value_type
                if type_str.startswith("typing"):
                    self.type_str_to_parameter[
                        type_str.replace("typing.", "")
                    ] = value_type
                if " " in type_str:
                    self.type_str_to_parameter[type_str.replace(" ", "")] = value_type
            if value_type.type_str_extras:
                for t in value_type.type_str_extras:
                    self.type_str_to_parameter[t] = value_type
        except Exception as ex:
            raise Exception("Failed to process %s: %s" % (value_type, ex))
        return value_type

    def get_value_type_of_obj(self, value, default=None):
        # type: (object, Optional[ValueType]) -> Optional[ValueType]
        # do we want to automatically parse str_list?
        # right now we do that, but as for obj_list
        # we can keep deterministic conversion only

        for item in self.discoverable_value_types:
            if item.is_type_of(value):
                return item
        return default

    def get_value_type_of_type(self, type_, inline_value_type=False):
        if isinstance(type_, ValueType):
            return type_
        elif isinstance(type_, type) and issubclass(type_, ValueType):
            return type_()
        return self._type_handler_from_type.get_value_type_of_type(
            type_=type_, inline_value_type=inline_value_type
        )

    def get_value_type_of_type_str(self, type_str):
        # type: (str) -> Optional[ValueType]
        return self._type_handler_from_type_str.get_value_type_of_type_str(type_str)

    def list_known_types(self):
        return list(self.type_str_to_parameter.keys())


def get_non_generic_type(type_):
    if isinstance(type_, type):
        return type_
    else:
        if hasattr(type_, "__extra__"):  # py 3.5/3.6
            return type_.__extra__
        elif hasattr(type_, "__origin__"):  # py 3.7
            return type_.__origin__
        else:
            return None


class TypeHandlerFromType(object):
    def __init__(self, registry):
        self.registry = registry  # type: ValueTypeRegistry

    def _get_value_type_of_type(self, type_, inline_value_type=False):
        if isinstance(
            type_, typing.TypeVar
        ):  # placeholder type for unparametrized generic type
            return None

        # from py3.7 issubclass(typing.List[str], typing.List) raises an exception, so we extract the simple type first
        non_generic_type = get_non_generic_type(type_)

        for value_type in self.registry.value_types:
            if value_type.type == type_:
                return value_type

            if non_generic_type is None:
                continue

            if isinstance(non_generic_type, type) and value_type.is_handler_of_type(
                non_generic_type
            ):
                return value_type
        if inline_value_type:
            return InlineValueType(type_)
        return None

    def get_value_type_of_type(self, type_, inline_value_type=False):
        lazy_load = None

        if is_Union(type_):
            union_params = get_Union_params(type_)
            if LazyLoad in union_params:
                lazy_load = True
            elif EagerLoad in union_params:
                lazy_load = False
            # take only first in Union
            type_ = union_params[0]
        if inspect.isclass(type_) and issubclass(type_, LazyLoad):
            lazy_load = True

        type_handler = self._get_value_type_of_type(
            type_, inline_value_type=inline_value_type
        )
        if not type_handler:
            return type_handler

        if isinstance(type_handler, _StructureValueType):
            if type_handler.type == typing.Dict:
                sub_type = self.get_dict_value_type(type_)
            else:
                sub_type = self.get_generic_item_type(type_)

            if sub_type:
                sub_type_handler = self.registry.get_value_type_of_type(
                    sub_type, inline_value_type=inline_value_type
                )
                type_handler = type_handler.with_sub_type_handler(sub_type_handler)
        if lazy_load is not None:
            type_handler = type_handler.with_lazy_load(lazy_load=lazy_load)
        return type_handler

    def get_dict_value_type(self, type_):
        args = _get_generic_params(type_)
        if args is not None and len(args) == 2:
            return args[1]
        return None

    def get_generic_item_type(self, type_):
        args = _get_generic_params(type_)
        if args is not None and len(args) == 1:
            return args[0]
        return None


def _get_generic_params(type_):
    if hasattr(type_, "__args__"):
        args = type_.__args__
        if isinstance(args, tuple) and args:
            return args
    return None


class TypeHandlerFromDocAnnotation(object):
    # re:   Optional[ Type1 ]
    _OPTIONAL_TYPE_RE = re.compile(r"^Optional\[(.+)\]$")
    # re:   Union[ Type1, .. ]
    _UNION_TYPE_RE = re.compile(r"^Union\[(.+)\]$")

    # re:   Union[ Type1, .. ]
    _SIMPLE_TYPE_RE = re.compile(r"^[^(\[]+\.(.+)$")
    # re:   Union[ Type1, .. ]
    _GENERIC_TYPE_RE = re.compile(r"^([^\[]+)\[(.+)\]$")

    _SPLIT_ARGS_RE = re.compile(r",(?![^()\[\]]*[)\]])")

    def __init__(self, registry):
        self.registry = registry  # type: ValueTypeRegistry

    def _get_value_type_of_type_str(self, type_str):
        r = self.registry
        options = [type_str, type_str.lower()]
        # let take only "last" part of   some_module.SomeType[]
        simple_type_match = self._SIMPLE_TYPE_RE.match(type_str)
        if simple_type_match:
            simple_type_str = simple_type_match.group(1)
            options += [simple_type_str, simple_type_str.lower()]

        options = [x[0] for x in itertools.groupby(options)]

        # we are looking for the exact match
        for ts in options:
            type_handler = r.type_str_to_parameter.get(ts)
            if type_handler:
                return type_handler

        # now we can try "generic"
        for ts in options:
            generic_match = self._GENERIC_TYPE_RE.match(ts)
            if not generic_match:
                continue
            main_type = generic_match.group(1)
            generic_part = generic_match.group(2)

            type_handler = r.type_str_to_parameter.get(main_type)
            if type_handler:
                if type_handler.type == typing.Dict:
                    parts = self._SPLIT_ARGS_RE.split(generic_part)
                    if len(parts) == 2:
                        generic_part = parts[1]
                    else:
                        continue

                sub_type_handler = r.get_value_type_of_type_str(generic_part)
                return type_handler.with_sub_type_handler(sub_type_handler)

        return None

    def get_value_type_of_type_str(self, type_str):
        # type: (str) -> Optional[ValueType]
        if not type_str:
            return None
        # if not REGEX_SIMPLE_TYPE.match(type_str):
        #     return None
        type_str = type_str.replace(" ", "")
        lazy_load = None
        # we remove Optional wrapping - we don't use that information
        optional_def = self._OPTIONAL_TYPE_RE.search(type_str)
        if optional_def:
            type_str = optional_def.group(1)

        union = self._UNION_TYPE_RE.search(type_str)
        if union:
            type_str = union.group(1)
            # we take only first argument on Union
            union_types = self._SPLIT_ARGS_RE.split(type_str)
            if "LazyLoad" in union_types:
                lazy_load = True
            type_str = union_types[0]

        if "DataList" in type_str:
            lazy_load = True

        type_handler = self._get_value_type_of_type_str(type_str)
        if not type_handler:
            return None

        if lazy_load is not None:
            type_handler = type_handler.with_lazy_load(lazy_load)

        return type_handler



# Copyright (c) 2016 Timo Furrer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# -*- coding: utf-8 -*-
"""
    Extension for the python ``click`` module to provide
    a group with a git-like *did-you-mean* feature.
"""

import difflib

from dbnd._vendor import click

__version__ = "0.0.3"


class DYMMixin(object):  # pylint: disable=too-few-public-methods
    """
    Mixin class for click MultiCommand inherited classes
    to provide git-like *did-you-mean* functionality when
    a certain command is not registered.
    """

    def __init__(self, *args, **kwargs):
        self.max_suggestions = kwargs.pop("max_suggestions", 3)
        self.cutoff = kwargs.pop("cutoff", 0.5)
        super(DYMMixin, self).__init__(*args, **kwargs)

    def resolve_command(self, ctx, args):
        """
        Overrides clicks ``resolve_command`` method
        and appends *Did you mean ...* suggestions
        to the raised exception message.
        """
        try:
            return super(DYMMixin, self).resolve_command(ctx, args)
        except click.exceptions.UsageError as error:
            error_msg = str(error)
            original_cmd_name = click.utils.make_str(args[0])
            matches = difflib.get_close_matches(original_cmd_name,
                                                self.list_commands(ctx), self.max_suggestions, self.cutoff)
            if matches:
                error_msg += '\n\nDid you mean one of these?\n    %s' % '\n    '.join(
                    matches)  # pylint: disable=line-too-long

            raise click.exceptions.UsageError(error_msg, error.ctx)


class DYMGroup(DYMMixin, click.Group):  # pylint: disable=too-many-public-methods
    """
    click Group to provide git-like
    *did-you-mean* functionality when a certain
    command is not found in the group.
    """


class DYMCommandCollection(DYMMixin, click.CommandCollection):  # pylint: disable=too-many-public-methods
    """
    click CommandCollection to provide git-like
    *did-you-mean* functionality when a certain
    command is not found in the group.
    """



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging

from dbnd import parameter
from dbnd._core.task_build.dbnd_decorator import task
from dbnd.tasks.doctor import system_airflow, system_dbnd, system_logging, system_python
from dbnd.tasks.doctor.doctor_report_builder import DoctorStatusReportBuilder


logger = logging.getLogger(__name__)


@task
def dbnd_doctor(
    python_sanity=parameter.value(True)[bool],
    airflow_sanity=parameter.value(True)[bool],
    logs=parameter.value(None)[bool],
    python_packages=parameter.value(None)[bool],
    check_time=datetime.datetime.now(),
    all=False,
):
    if all:
        # change only "none" params
        logs = True if logs is None else logs
        python_packages = True if python_packages is None else python_packages

    main_report = DoctorStatusReportBuilder("Dbnd Doctor")
    main_report.log("check_time", check_time)

    system_report = system_dbnd.dbnd_status()
    logger.debug("system_report: %s", system_report)

    if python_sanity:
        system_python_report = system_python.python_status(
            python_packages=python_packages
        )
        main_report.add_sub_report(system_python_report)
    if airflow_sanity:
        airflow_report = system_airflow.airflow_status()
        main_report.add_sub_report(airflow_report)
    if logs:
        system_logging_report = system_logging.logging_status()
        main_report.add_sub_report(system_logging_report)

    logger.info("Your system is good to go! Enjoy Databand!")
    return main_report.get_status_str()



# © Copyright Databand.ai, an IBM Company 2022


def is_instance_by_class_name(obj, class_name):
    # If it's metaclass
    if issubclass(obj.__class__, type):
        return obj.__class__.__name__ == class_name

    for cls in obj.__class__.mro():
        if cls.__name__ == class_name:
            return True
    return False



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import absolute_import

import logging
import os

from dbnd._vendor.dulwich import porcelain
from dbnd._vendor.dulwich.repo import Repo


logger = logging.getLogger(__name__)

GIT_ENV = "DBND_PROJECT_GIT_COMMIT"


def is_git_dirty(path=None, verbose=False):
    try:
        repo = Repo.discover(path)
        status = porcelain.status(repo.path)
        is_dirty = any(
            [
                status.staged["add"],
                status.staged["delete"],
                status.staged["modify"],
                len(status.unstaged),
                len(status.untracked),
            ]
        )
        return is_dirty
    except Exception as ex:
        if verbose:
            logger.warning("Failed to get GIT status %s: %s", path, ex)
        return None


def get_git_commit(path, verbose=False):
    env_commit = os.environ.get(GIT_ENV)
    if env_commit:
        return env_commit
    try:
        repo = Repo.discover(path)
        commit = repo.head().decode("utf-8")
        return commit
    except Exception as ex:
        if verbose:
            logger.warning("Failed to get GIT version of %s: %s", path, ex)
        return None



# © Copyright Databand.ai, an IBM Company 2022

import json
import logging
import typing

from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from pandas.core.dtypes.common import is_bool_dtype, is_numeric_dtype, is_string_dtype

from dbnd._core.tracking.schemas.column_stats import (
    ColumnStatsArgs,
    get_column_stats_by_col_name,
)


if typing.TYPE_CHECKING:
    from dbnd._core.tracking.log_data_request import LogDataRequest
    from targets.value_meta import ValueMetaConf

logger = logging.getLogger(__name__)


class PandasHistograms(object):
    """
    calculates histograms and stats on pandas dataframe.
    """

    def __init__(self, df, meta_conf):
        # type: (pd.DataFrame, ValueMetaConf) -> None
        self.df = df
        self.meta_conf = meta_conf

    def get_histograms_and_stats(
        self,
    ) -> Tuple[List[ColumnStatsArgs], Dict[str, List[List]]]:
        columns_stats, histograms = [], {}

        if self.meta_conf.log_stats:
            columns_stats = self._calculate_stats(self.df)

        if self.meta_conf.log_histograms:
            hist_column_names = self._get_column_names_from_request(
                self.df, self.meta_conf.log_histograms
            )
            df_histograms = self.df.filter(hist_column_names)
            if pd.__version__ > "1":
                # return_type="expand" and orient="list" are used to stabilize produced
                # histograms data structure across pandas v0 & v1
                histograms = df_histograms.apply(
                    self._calculate_histograms,
                    result_type="expand",
                    args=(columns_stats,),
                ).to_dict(orient="list")
            else:
                histograms = df_histograms.apply(
                    self._calculate_histograms, args=(columns_stats,)
                ).to_dict()
                histograms = {k: list(v) for k, v in histograms.items()}

        return columns_stats, histograms

    def _get_column_names_from_request(self, df, data_request):
        # type: (pd.DataFrame, LogDataRequest) -> List[str]
        column_names = list(data_request.include_columns)
        for column_name, column_type in df.dtypes.items():
            if data_request.include_all_string and is_string_dtype(column_type):
                column_names.append(column_name)
            elif data_request.include_all_boolean and is_bool_dtype(column_type):
                column_names.append(column_name)
            elif data_request.include_all_numeric and is_numeric_dtype(column_type):
                column_names.append(column_name)

        if len(data_request.exclude_columns) > 0:
            if len(column_names) == 0 and len(df.dtypes.keys()) > 0:
                column_names.extend(df.dtypes.keys())

            column_names = [
                column
                for column in column_names
                if column not in data_request.exclude_columns
            ]

        return column_names

    def _calculate_stats(self, df):
        # type: (pd.DataFrame) -> Dict[str, Dict]
        stats_column_names = self._get_column_names_from_request(
            df, self.meta_conf.log_stats
        )
        df = df.filter(stats_column_names)
        try:
            stats = df.describe(include="all").to_json()
        except Exception as e:
            logger.warning("Failed to describe df: %s", e)
            stats = df.explode().describe(include="all").to_json()
        stats = json.loads(stats)
        stats = self._remove_none_values(stats)
        columns_stats: List[ColumnStatsArgs] = []
        for column_name in stats.keys():
            # Calculate ColumnStatsArgs metrics
            records_count = df[column_name].size
            null_count = np.count_nonzero(pd.isnull(df[column_name]))
            try:
                distinct_count = len(df[column_name].unique())
            except Exception:
                logger.warning("Failed to determine column type for: %s.", column_name)
                try:
                    distinct_count = len(df[column_name].astype("str").unique())
                except Exception:
                    # Support pandas >= v1.0
                    distinct_count = len(df[column_name].astype("string").unique())

            column_type = self._get_column_type(df[column_name])
            column_stats = ColumnStatsArgs(
                column_name=column_name,
                column_type=column_type,
                records_count=records_count,
                null_count=null_count,
                distinct_count=distinct_count,
                most_freq_value=stats[column_name].get("top"),
                most_freq_value_count=stats[column_name].get("freq"),
                unique_count=stats[column_name].get("unique"),
                mean_value=stats[column_name].get("mean"),
                min_value=stats[column_name].get("min"),
                max_value=stats[column_name].get("max"),
                std_value=stats[column_name].get("std"),
                quartile_1=stats[column_name].get("25%"),
                quartile_2=stats[column_name].get("50%"),
                quartile_3=stats[column_name].get("75%"),
            )
            columns_stats.append(column_stats)

        return columns_stats

    def _remove_none_values(self, input_dict):
        """remove none values from dict recursively"""
        for key, value in list(input_dict.items()):
            if value is None:
                input_dict.pop(key)
            elif isinstance(value, dict):
                self._remove_none_values(value)
        return input_dict

    def _get_column_type(self, column):
        # type: (pd.Series) -> str
        first_index = column.first_valid_index()
        if first_index is None:
            return column.dtype.name
        first_value = column.at[first_index]
        return type(first_value).__name__

    def _calculate_histograms(self, df_column, columns_stats):
        # type: (pd.Series, List[ColumnStatsArgs]) -> Optional[Tuple[List, List]]
        try:
            if len(df_column) == 0:
                return

            column_type = self._get_column_type(df_column)
            if "1" < pd.__version__ < "1.2":
                # handle  'Float32', 'Float64', 'Int8', 'Int16', 'Int32', 'Int64', 'UInt8', 'UInt16', 'UInt32', 'UInt64'
                column_type = column_type.lower()
            if is_string_dtype(column_type) or is_bool_dtype(column_type):
                counts = df_column.value_counts()  # type: pd.Series
                column_stats = get_column_stats_by_col_name(
                    columns_stats, df_column.name
                )
                if column_stats and column_stats.null_count:
                    null_column = pd.Series([column_stats.null_count], index=[None])
                    counts = counts.append(null_column)
                    counts = counts.sort_values(ascending=False)
                if len(counts) > 50:
                    counts, tail = counts[:49], counts[49:]
                    tail_sum = pd.Series([tail.sum()], index=["_others"])
                    counts = counts.append(tail_sum)
                values = counts.index
            elif is_numeric_dtype(column_type):
                df_column = df_column.dropna()
                counts, values = np.histogram(
                    df_column, bins=20
                )  # Tuple[np.array, np.array]
            else:
                return

            return counts.tolist(), values.tolist()
        except Exception:
            logger.exception(
                "log_histogram: Something went wrong for column '%s'", df_column.name
            )



# © Copyright Databand.ai, an IBM Company 2022

import re

from typing import Callable, Optional

import more_itertools


def camel_to_snake(name, placeholder="_"):
    # type: (str, str) -> str
    # Convert argument names from lowerCamelCase to snake case.
    # "AbBcDe" -> "ab_bc_de"
    return re.sub(
        r"[A-Z]",
        lambda x: (
            placeholder if x.start(0) > 0 else ""
        )  # don't add to the first character
        + x.group(0).lower(),
        name,
    )


def clean_job_name(
    value, enabled_characters=r"\-_", placeholder="_", max_size=None, postfix=None
):
    # type:(str,str,str, int, Optional[str]) -> str
    """
    @param value: the base value to transform
    @param enabled_characters: all the allowed characters beside alphanumeric [white list]
    @param placeholder: the char that would be used to replace any character that is not alphanumeric or in `enabled_characters`
    @param max_size: the maximum length allowed for the output
    @param postfix: optional string to add to end of the result value
    @return: a transformation of the value, by all the configurations
    """
    value = camel_to_snake(value, placeholder=placeholder)
    enabled_characters = re.escape(enabled_characters)
    # clean all garbage
    value = re.sub(r"[^a-z0-9%s]" % enabled_characters, placeholder, value)

    # clean all duplicated special charaters:  .-  or -- or ___
    value = re.sub(
        r"([{enabled_characters}])[{enabled_characters}]+".format(
            enabled_characters=enabled_characters
        ),
        r"\1",
        value,
    )
    if max_size:
        if postfix:
            max_size -= len(postfix)
        value = value[:max_size]

    if postfix:
        postfix = clean_job_name(
            value=postfix,
            enabled_characters=enabled_characters,
            placeholder=placeholder,
        )
        value += postfix
        # different from the first replace, we are replacing using second character
        value = re.sub(
            r"[{enabled_characters}]+([{enabled_characters}])".format(
                enabled_characters=enabled_characters
            ),
            r"\1",
            value,
        )
    return value


def str_or_none(value):
    if value is None:
        return None
    return str(value)


def safe_short_string(value, max_value_len=1000, tail=False):
    """Returns the string limited by max_value_len parameter.

    Parameters:
        value (str): the string to be shortened
        max_value_len (int): max len of output
        tail (bool):
    Returns:
        str: the string limited by max_value_len parameter

    >>> safe_short_string('abcdefghijklmnopqrstuvwxyz', max_value_len=20)
    'abcdef... (6 of 26)'
    >>> safe_short_string('abcdefghijklmnopqrstuvwxyz', max_value_len=20, tail=True)
    '(6 of 26) ...uvwxyz'
    >>> (safe_short_string(''), safe_short_string(None))
    ('', None)
    >>> safe_short_string('qwerty', max_value_len=4)
    '... (0 of 6)'
    >>> safe_short_string('qwerty', max_value_len=-4)
    '... (0 of 6)'
    >>> safe_short_string('qwerty'*123, max_value_len=0)
    '... (0 of 738)'
    >>> safe_short_string('qwerty', max_value_len=4, tail=True)
    '(0 of 6) ...'
    >>> safe_short_string('qwerty', max_value_len=-4, tail=True)
    '(0 of 6) ...'
    >>> safe_short_string('qwerty'*123, max_value_len=0, tail=True)
    '(0 of 738) ...'
    >>> safe_short_string('qwerty', max_value_len='exception')
    "ERROR: Failed to shorten string: '>' not supported between instances of 'int' and 'str'"
    """
    try:
        if not value:
            return value
        if len(value) > max_value_len:
            placeholder = "... (%s of %s)".format(max_value_len, len(value))
            actual_len = max_value_len - len(placeholder)
            actual_len = 0 if actual_len < 0 else actual_len
            if tail:
                value = "(%s of %s) ...%s" % (
                    actual_len,
                    len(value),
                    value[len(value) - actual_len :],
                )
            else:
                value = "%s... (%s of %s)" % (
                    value[:actual_len],
                    actual_len,
                    len(value),
                )
        return value
    except Exception as ex:
        # we don't want to fail here
        return "ERROR: Failed to shorten string: %s" % ex


def pluralize(s, n, plural_form=None):
    if n == 1:
        return s
    else:
        return plural_form or s + "s"


def strip_whitespace(string):
    return re.sub(r"\s+", " ", string.strip(), flags=re.UNICODE)


def is_task_name_driver(task_name):
    return "dbnd_driver" in task_name


# Regex to catch start of dbnd logs (they're starting like this: [2020-05-07 17:47:07,768])
DBND_LOGS_REGEX = re.compile(r"\[\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2},\d{3}]")


def merge_dbnd_and_spark_logs(dbnd, spark):
    """Merges dbnd and spark logs into single array and sorts them by timestamp provided.
    dbnd logs are multi-lined so we have to perform a tricky comparison and append here.
    """
    result = []

    dbnd_idx = 0
    spark_idx = 0

    while True:
        dbnd_line = None if dbnd_idx >= len(dbnd) else dbnd[dbnd_idx]
        spark_line = None if spark_idx >= len(spark) else spark[spark_idx]

        if spark_line is None and dbnd_line is None:
            break

        if spark_line is None:
            result.append(dbnd_line)
            dbnd_idx += 1
            continue

        if dbnd_line is None:
            result.append(spark_line)
            spark_idx += 1
            continue

        # dbnd logs are multi-lined and the first line is timestamp starting from [.
        # However, the rest of lines in the log record can also start with '[' so we have to check the line with regex
        if dbnd_line.startswith("[") and DBND_LOGS_REGEX.match(dbnd_line):
            if dbnd_line < spark_line:
                result.append(dbnd_line)
                dbnd_idx += 1
            else:
                result.append(spark_line)
                spark_idx += 1
        else:
            result.append(dbnd_line)
            dbnd_idx += 1

    return result


def strip_by(predicate, string):
    # type: (Callable[[str], bool], str) -> str
    """
    striping any characters in the beginning or the end matching the predicate
    >>> strip_by(lambda c: not c.isalnum(), "...123123.123213.123..asd.22..") == "123123.123213.123..asd.22"
    """
    return "".join(more_itertools.strip(string, predicate))


def truncate_msg(msg, head_size, tail_size):
    msg_size = len(msg)
    if msg_size <= head_size + tail_size:
        return msg

    if tail_size <= 0 and head_size <= 0:
        return ""

    truncated_msg_head = f"{msg[:head_size]}... (first {head_size} of {len(msg)})"
    if tail_size <= 0:
        return truncated_msg_head

    truncated_msg_tail = f"(last {tail_size} of {len(msg)}) ...{msg[-tail_size:]}"
    if head_size <= 0:
        return truncated_msg_tail

    return f"{truncated_msg_head}\n{truncated_msg_tail}"



# © Copyright Databand.ai, an IBM Company 2022

import logging
import typing

from typing import Any, Dict, List, Optional, Union

import attr
import six

from dbnd._core.constants import (
    DbndDatasetOperationType,
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    MetricSource,
)
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.log.external_exception_logging import (
    capture_tracking_exception,
    log_exception_to_server,
)
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.settings.tracking_config import get_value_meta
from dbnd._core.task_run.task_run_ctrl import TaskRunCtrl
from dbnd._core.tracking.log_data_request import LogDataRequest
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.timezone import utcnow
from targets import Target
from targets.value_meta import ValueMeta, ValueMetaConf


if typing.TYPE_CHECKING:
    from datetime import datetime

    import pandas as pd
    import pyspark.sql as spark

    from dbnd._core.tracking.backends import TrackingStore
    from dbnd_postgres.postgres_values import PostgresTable
    from dbnd_snowflake.snowflake_values import SnowflakeTable

logger = logging.getLogger(__name__)


class TaskRunTracker(TaskRunCtrl):
    def __init__(self, task_run, tracking_store):
        super(TaskRunTracker, self).__init__(task_run=task_run)
        self.tracking_store = tracking_store  # type: TrackingStore

    def task_run_url(self):
        run_tracker = self.run.tracker
        if not run_tracker.databand_url:
            return None

        return "{databand_url}/app/jobs/{job_name}/{run_uid}/{task_run_uid}".format(
            databand_url=run_tracker.databand_url,
            job_name=self.run.job_name,
            run_uid=self.run.run_uid,
            task_run_uid=self.task_run_uid,
        )

    # Task Handlers
    def save_task_run_log(self, log_preview, local_log_path=None):
        self.tracking_store.save_task_run_log(
            task_run=self.task_run, log_body=log_preview, local_log_path=local_log_path
        )

    def log_parameter_data(
        self, parameter, target, value, operation_type, operation_status
    ):
        # type: (TaskRunTracker, ParameterDefinition, Target, Any, DbndTargetOperationType, DbndTargetOperationStatus) -> None
        tracking_conf = self.settings.tracking
        if not tracking_conf.log_value_meta or value is None:
            return

        try:
            target.target_meta = get_value_meta(
                value,
                parameter.value_meta_conf,
                tracking_config=tracking_conf,
                value_type=parameter.value_type,
                target=target,
            )
            # FIXME If we failed to get target meta for some reason, target operation won't be logged!
            if target.target_meta is None:
                return

            self.tracking_store.log_target(
                task_run=self.task_run,
                target=target,
                target_meta=target.target_meta,
                operation_type=operation_type,
                operation_status=operation_status,
                param_name=parameter.name,
                task_def_uid=parameter.task_definition_uid,
            )
        except Exception as ex:
            log_exception(
                "Error occurred during target logging for %s" % (target,),
                ex,
                non_critical=True,
            )

    def _log_metrics(self, metrics):
        # type: (List[Metric]) -> None
        return self.tracking_store.log_metrics(task_run=self.task_run, metrics=metrics)

    def log_dbt_metadata(self, dbt_metadata):
        self.tracking_store.log_dbt_metadata(
            dbt_run_metadata=dbt_metadata, task_run=self.task_run
        )

    @capture_tracking_exception
    def log_metrics(self, metrics_dict, source=None, timestamp=None):
        # type: (Dict[str, Any], Optional[str], Optional[datetime]) -> None
        """
        Logs all the metrics in the metrics dict to the tracker.
        @param metrics_dict: name-value pairs of metrics to log
        @param source: optional name of the metrics source
        @param timestamp: optional timestamp of the metrics
        """
        metrics = [
            Metric(key=key, value=value, source=source, timestamp=timestamp or utcnow())
            for key, value in six.iteritems(metrics_dict)
        ]

        return self.tracking_store.log_metrics(task_run=self.task_run, metrics=metrics)

    def log_artifact(self, name, artifact):
        try:
            # file storage will save file
            # db will save path
            artifact_target = self.task_run.meta_files.get_artifact_target(name)
            self.tracking_store.log_artifact(
                task_run=self.task_run,
                name=name,
                artifact=artifact,
                artifact_target=artifact_target,
            )
        except Exception as ex:
            log_exception(
                "Error occurred during log_artifact for %s" % (name,),
                ex,
                non_critical=True,
            )

    def log_metric(self, key, value, timestamp=None, source=None):
        # type: (str, Any, Optional[datetime], Optional[MetricSource]) -> None
        try:
            self.log_metrics({key: value}, source, timestamp)
        except Exception as ex:
            log_exception(
                "Error occurred during log_metric for %s" % (key,),
                ex,
                non_critical=True,
            )

    def log_data(
        self,
        key,  # type: str
        data,  # type: Union[pd.DataFrame, spark.DataFrame, PostgresTable, SnowflakeTable]
        meta_conf,  # type: ValueMetaConf
        path=None,  # type: Optional[Union[Target,str]]
        operation_type=DbndTargetOperationType.read,  # type: DbndTargetOperationType
        operation_status=DbndTargetOperationStatus.OK,  # type: DbndTargetOperationStatus
        raise_on_error=False,  # type: bool
    ):  # type: (...) -> None
        try:
            # Combine meta_conf with the config settings
            value_meta = get_value_meta(
                data, meta_conf, tracking_config=self.settings.tracking
            )
            if not value_meta:
                logger.warning(
                    "Couldn't log the wanted data {name}, reason - can't log objects of type {value_type} ".format(
                        name=key, value_type=type(data)
                    )
                )
                return

            ts = utcnow()

            if path:
                self.tracking_store.log_target(
                    task_run=self.task_run,
                    target=path,
                    target_meta=value_meta,
                    operation_type=operation_type,
                    operation_status=operation_status,
                    param_name=key,
                )

            self.log_value_metrics(key, meta_conf, value_meta, ts)

        except Exception as ex:
            log_exception(
                "Error occurred during log_dataframe for %s" % (key,),
                ex,
                non_critical=not raise_on_error,
            )
            if raise_on_error:
                raise

    def log_value_metrics(self, key, meta_conf, value_meta, ts=None):
        ts = ts or utcnow()
        metrics = value_meta.build_metrics_for_key(key, meta_conf)
        if metrics["user"]:
            self._log_metrics(metrics["user"])

        if metrics["histograms"]:
            self.tracking_store.log_histograms(
                task_run=self.task_run, key=key, value_meta=value_meta, timestamp=ts
            )
        if not (metrics["user"] or metrics["histograms"]):
            logger.info("No metrics to log_data(key={})".format(key))

    def log_dataset(self, op_report):
        # type: (DatasetOperationReport) -> None
        data_meta = self._calc_meta_data(op_report.data, op_report.meta_conf)

        records, columns = data_meta.data_dimensions or (None, None)
        # if the row count or column count exist,
        # we need to override the match field(row/column) in data dimension
        # if one of them not exist we need to leave it with the original value
        # examples:
        # row_count=4, data_dimensions=(777,5) => data_dimensions=(4,5)
        # column_count=2, data_dimensions=(777,5) => data_dimensions=(777,2)
        if op_report.row_count or op_report.column_count:
            data_meta.data_dimensions = (
                op_report.row_count or records,
                op_report.column_count or columns,
            )
        self.tracking_store.log_dataset(
            task_run=self.task_run,
            operation_path=op_report.op_path,
            data_meta=data_meta,
            operation_type=op_report.op_type,
            operation_status=op_report.status,
            operation_error=op_report.error,
            operation_source=data_meta.op_source
            if data_meta.op_source
            else op_report.operation_source,
            with_partition=op_report.with_partition,
        )

        if op_report.send_metrics:
            dataset_name = _get_dataset_name(
                op_report.op_path, op_report.with_partition
            )

            self.log_value_metrics(dataset_name, op_report.meta_conf, data_meta)

    def _calc_meta_data(self, data, meta_conf):
        # type: (Any, ValueMetaConf) -> ValueMeta
        data_meta = None
        if data is not None and meta_conf is not None:
            # Combine meta_conf with the config settings
            try:
                data_meta = get_value_meta(
                    data, meta_conf, tracking_config=self.settings.tracking
                )
            except Exception as e:
                log_exception_to_server(e)

        if data_meta is None:
            data_meta = ValueMeta("")

        return data_meta


def _get_dataset_name(operation_path, with_partition):
    """
    This is a temporary patch to report metrics with log_dataset.
    Should be removed in favor of reporting the metrics with the dataset operation
    """
    # todo: deprecate the name calculation from the sdk, move only to backend
    try:
        from urllib.parse import urlparse
    except ImportError:
        from urlparse import urlparse

    import itertools
    import os

    path = urlparse(str(operation_path)).path  # type: str

    if with_partition:
        dirname, basename = os.path.split(path)
        if "." in basename and not path.endswith("/"):
            path = dirname

    path_parts = path.strip("/").split("/")
    if with_partition:
        path_parts = list(itertools.takewhile(lambda part: "=" not in part, path_parts))

    dataset_name = ".".join(path_parts)

    # we might not successfully extract a name from the path so as a
    # fallback we return the original value
    if dataset_name == "":
        return operation_path

    return dataset_name


@attr.s(slots=True, frozen=False, kw_only=True)
class DatasetOperationReport(object):
    """
    Holds the information about an operation to report
    """

    op_path: Union[Target, str] = attr.ib()
    op_type: Union[DbndDatasetOperationType, str] = attr.ib(
        # convert str of "read" or "write" to operation type
        converter=lambda op_type: DbndDatasetOperationType[op_type]
        if isinstance(op_type, str)
        else op_type
    )
    with_preview: bool = attr.ib()
    with_schema: bool = attr.ib()

    data: Optional[Any] = attr.ib()
    with_histograms: Optional[Union[bool, str, List[str], LogDataRequest]] = attr.ib()
    with_stats: Optional[Union[bool, str, List[str], LogDataRequest]] = attr.ib()

    success: bool = attr.ib(default=True)
    error: Optional[str] = attr.ib(default=None)
    send_metrics = attr.ib(default=None)
    with_partition = attr.ib(default=None)
    row_count = attr.ib(default=None)
    column_count = attr.ib(default=None)

    operation_source: Optional[str] = attr.ib(default=None)

    def set(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                logger.warning(
                    "Can't set attribute {} for DatasetOperationLogger, no such attribute".format(
                        k
                    )
                )

    def set_data(self, data):
        self.set(data=data)

    def set_error(self, error):
        # type: (str) -> None
        self.set(success=False, error=error)

    def set_success(self):
        self.set(success=True)

    @property
    def meta_conf(self):
        return ValueMetaConf(
            log_preview=self.with_preview,
            log_schema=self.with_schema,
            log_size=self.with_schema,
            log_stats=self.with_stats,
            log_histograms=self.with_histograms,
        )

    @property
    def status(self):
        return (
            DbndTargetOperationStatus.OK
            if self.success
            else DbndTargetOperationStatus.NOK
        )



# -*- coding: utf-8 -*-
#
# Copyright 2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# This file has been modified by databand.ai to support dbnd orchestration and tracking.


import io
import locale
import os
import re
import warnings

import six

from targets.pipes.base import BaseWrapper, WrappedFormat


class NewlineWrapper(BaseWrapper):
    def __init__(self, stream, newline=None):
        if newline is None:
            self.newline = newline
        else:
            self.newline = newline.encode("ascii")

        if self.newline not in (b"", b"\r\n", b"\n", b"\r", None):
            raise ValueError(
                "newline need to be one of {b'', b'\r\n', b'\n', b'\r', None}"
            )
        super(NewlineWrapper, self).__init__(stream)

    def read(self, n=-1):
        b = self._stream.read(n)

        if self.newline == b"":
            return b

        if self.newline is None:
            newline = b"\n"

        return re.sub(b"(\n|\r\n|\r)", newline, b)

    def writelines(self, lines):
        if self.newline is None or self.newline == "":
            newline = os.linesep.encode("ascii")
        else:
            newline = self.newline

        self._stream.writelines(
            (re.sub(b"(\n|\r\n|\r)", newline, line) for line in lines)
        )

    def write(self, b):
        if self.newline is None or self.newline == "":
            newline = os.linesep.encode("ascii")
        else:
            newline = self.newline

        self._stream.write(re.sub(b"(\n|\r\n|\r)", newline, b))


class TextWrapper(io.TextIOWrapper):
    def __exit__(self, *args):
        # io.TextIOWrapper close the file on __exit__, let the underlying file decide
        if not self.closed and self.writable():
            super(TextWrapper, self).flush()

        self._stream.__exit__(*args)

    def __del__(self, *args):
        # io.TextIOWrapper close the file on __del__, let the underlying file decide
        if not self.closed and self.writable():
            super(TextWrapper, self).flush()

        try:
            self._stream.__del__(*args)
        except AttributeError:
            pass

    def __init__(self, stream, *args, **kwargs):
        self._stream = stream
        try:
            super(TextWrapper, self).__init__(stream, *args, **kwargs)
        except TypeError:
            pass

    def __getattr__(self, name):
        if name == "_stream":
            raise AttributeError(name)
        return getattr(self._stream, name)

    def __enter__(self):
        self._stream.__enter__()
        return self


class MixedUnicodeBytesWrapper(BaseWrapper):
    """ """

    def __init__(self, stream, encoding=None):
        if encoding is None:
            encoding = locale.getpreferredencoding()
        self.encoding = encoding
        super(MixedUnicodeBytesWrapper, self).__init__(stream)

    def write(self, b):
        self._stream.write(self._convert(b))

    def writelines(self, lines):
        self._stream.writelines((self._convert(line) for line in lines))

    def _convert(self, b):
        if isinstance(b, six.text_type):
            b = b.encode(self.encoding)
            warnings.warn("Writing unicode to byte stream", stacklevel=2)
        return b


class TextPipeline(WrappedFormat):

    input = "unicode"
    output = "bytes"
    wrapper_cls = TextWrapper


class MixedUnicodeBytesFormat(WrappedFormat):

    output = "bytes"
    wrapper_cls = MixedUnicodeBytesWrapper


class NewlinePipeline(WrappedFormat):

    input = "bytes"
    output = "bytes"
    wrapper_cls = NewlineWrapper



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.constants import _TaskParamContainer
from dbnd._core.current import get_settings
from dbnd._core.task import Config
from dbnd._core.task_build.task_registry import (
    build_task_from_config,
    get_task_registry,
)
from targets.values import ValueType


class TaskValueType(ValueType):
    """
    A value that takes another databand task class.

    When used programatically, the parameter should be specified
    directly with the :py:class:`dbnd.tasks.Task` (sub) class. Like
    ``MyMetaTask(my_task_param=my_tasks.MyTask)``. On the command line,
    you specify the :py:meth:`dbnd.tasks.Task.get_task_family`. Like

    .. code-block:: console

            $ dbnd --module my_tasks MyMetaTask --my_task_param my_namespace.MyTask

    Where ``my_namespace.MyTask`` is defined in the ``my_tasks`` python module.

    When the :py:class:`dbnd.tasks.Task` class is instantiated to an object.
    The value will always be a task class (and not a string).
    """

    type = _TaskParamContainer

    def parse_from_str(self, input):
        """
        Parse a task_famly using the :class:`~dbnd._core.register.Register`
        """

        task_cls = get_task_registry().get_task_cls(input)
        return task_cls()

    def to_str(self, cls):
        """
        Converts the :py:class:`dbnd.tasks.Task` (sub) class to its family name.
        """
        return cls.get_task_family()

    def is_type_of(self, value):
        return isinstance(value, _TaskParamContainer)


class ConfigValueType(ValueType):
    type = Config
    support_from_str = True

    def __init__(self, config_cls):
        self.config_cls = config_cls

    def parse_from_str(self, input):
        """
        Parse a task_famly using the :class:`~dbnd._core.register.Register`
        """

        from dbnd._core.settings.env import EnvConfig

        if isinstance(self.config_cls, EnvConfig):
            return get_settings().get_env_config(input)
        return build_task_from_config(input, expected_type=self.config_cls)

    def to_str(self, x):
        """
        Converts the :py:class:`dbnd.tasks.Task` (sub) class to its family name.
        """
        if x:
            return x.task_name
        return super(ConfigValueType, self).to_str(x)

    def is_type_of(self, value):
        return isinstance(value, Config)

    def to_repr(self, x):
        if x:
            return '"%s"' % x.task_name
        return "None"



# coding: utf-8
# Vendorized zipp package
# License: MIT License

# Copyright Jason R. Coombs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import division

import io
import sys
import posixpath
import zipfile
import functools
import itertools

import more_itertools

__metaclass__ = type


def _parents(path):
    """
    Given a path with elements separated by
    posixpath.sep, generate all parents of that path.
    >>> list(_parents('b/d'))
    ['b']
    >>> list(_parents('/b/d/'))
    ['/b']
    >>> list(_parents('b/d/f/'))
    ['b/d', 'b']
    >>> list(_parents('b'))
    []
    >>> list(_parents(''))
    []
    """
    return itertools.islice(_ancestry(path), 1, None)


def _ancestry(path):
    """
    Given a path with elements separated by
    posixpath.sep, generate all elements of that path
    >>> list(_ancestry('b/d'))
    ['b/d', 'b']
    >>> list(_ancestry('/b/d/'))
    ['/b/d', '/b']
    >>> list(_ancestry('b/d/f/'))
    ['b/d/f', 'b/d', 'b']
    >>> list(_ancestry('b'))
    ['b']
    >>> list(_ancestry(''))
    []
    """
    path = path.rstrip(posixpath.sep)
    while path and path != posixpath.sep:
        yield path
        path, tail = posixpath.split(path)


class Path:
    """
    A pathlib-compatible interface for zip files.
    Consider a zip file with this structure::
        .
        ├── a.txt
        └── b
            ├── c.txt
            └── d
                └── e.txt
    >>> data = io.BytesIO()
    >>> zf = zipfile.ZipFile(data, 'w')
    >>> zf.writestr('a.txt', 'content of a')
    >>> zf.writestr('b/c.txt', 'content of c')
    >>> zf.writestr('b/d/e.txt', 'content of e')
    >>> zf.filename = 'abcde.zip'
    Path accepts the zipfile object itself or a filename
    >>> root = Path(zf)
    From there, several path operations are available.
    Directory iteration (including the zip file itself):
    >>> a, b = root.iterdir()
    >>> a
    Path('abcde.zip', 'a.txt')
    >>> b
    Path('abcde.zip', 'b/')
    name property:
    >>> b.name
    'b'
    join with divide operator:
    >>> c = b / 'c.txt'
    >>> c
    Path('abcde.zip', 'b/c.txt')
    >>> c.name
    'c.txt'
    Read text:
    >>> c.read_text()
    'content of c'
    existence:
    >>> c.exists()
    True
    >>> (b / 'missing.txt').exists()
    False
    Coercion to string:
    >>> str(c)
    'abcde.zip/b/c.txt'
    """

    __repr = "{self.__class__.__name__}({self.root.filename!r}, {self.at!r})"

    def __init__(self, root, at=""):
        self.root = (
            root
            if isinstance(root, zipfile.ZipFile)
            else zipfile.ZipFile(self._pathlib_compat(root))
        )
        self.at = at

    @staticmethod
    def _pathlib_compat(path):
        """
        For path-like objects, convert to a filename for compatibility
        on Python 3.6.1 and earlier.
        """
        try:
            return path.__fspath__()
        except AttributeError:
            return str(path)

    @property
    def open(self):
        return functools.partial(self.root.open, self.at)

    @property
    def name(self):
        return posixpath.basename(self.at.rstrip("/"))

    def read_text(self, *args, **kwargs):
        with self.open() as strm:
            return io.TextIOWrapper(strm, *args, **kwargs).read()

    def read_bytes(self):
        with self.open() as strm:
            return strm.read()

    def _is_child(self, path):
        return posixpath.dirname(path.at.rstrip("/")) == self.at.rstrip("/")

    def _next(self, at):
        return Path(self.root, at)

    def is_dir(self):
        return not self.at or self.at.endswith("/")

    def is_file(self):
        return not self.is_dir()

    def exists(self):
        return self.at in self._names()

    def iterdir(self):
        if not self.is_dir():
            raise ValueError("Can't listdir a file")
        subs = map(self._next, self._names())
        return filter(self._is_child, subs)

    def __str__(self):
        return posixpath.join(self.root.filename, self.at)

    def __repr__(self):
        return self.__repr.format(self=self)

    def joinpath(self, add):
        add = self._pathlib_compat(add)
        next = posixpath.join(self.at, add)
        next_dir = posixpath.join(self.at, add, "")
        names = self._names()
        return self._next(next_dir if next not in names and next_dir in names else next)

    __truediv__ = joinpath

    @staticmethod
    def _implied_dirs(names):
        return more_itertools.unique_everseen(
            parent + "/"
            for name in names
            for parent in _parents(name)
            if parent + "/" not in names
        )

    @classmethod
    def _add_implied_dirs(cls, names):
        return names + list(cls._implied_dirs(names))

    @property
    def parent(self):
        parent_at = posixpath.dirname(self.at.rstrip('/'))
        if parent_at:
            parent_at += '/'
        return self._next(parent_at)

    def _names(self):
        return self._add_implied_dirs(self.root.namelist())

    if sys.version_info < (3,):
        __div__ = __truediv__



# © Copyright Databand.ai, an IBM Company 2022

from threading import RLock

from dbnd._core.utils.seven import get_ident
from dbnd._vendor.cachetools import cached as cachetools_cached, keys as cachetools_keys
from dbnd._vendor.cachetools.keys import _HashedTuple, _kwmark


class memoized_property(object):
    """A read-only @property that is only evaluated once."""

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


def cached(cache=None, key=cachetools_keys.hashkey, lock=None):
    if cache is None:
        cache = {}
    return cachetools_cached(cache=cache, key=key, lock=lock)


def _thread_safe_hashkey(*args, **kwargs):
    """Return a cache key for the specified hashable arguments."""
    thread_id = get_ident()
    if kwargs:
        return _HashedTuple(args + (thread_id,) + sum(sorted(kwargs.items()), _kwmark))
    else:
        return _HashedTuple(args + (thread_id,))


_per_thread_safe_lock = RLock()


def per_thread_cached(cache=None):
    if cache is None:
        cache = {}
    return cachetools_cached(
        cache=cache, key=_thread_safe_hashkey, lock=_per_thread_safe_lock
    )



# © Copyright Databand.ai, an IBM Company 2022

from typing import Any, List, Optional

import attr

from more_itertools import first


class ConfigValuePriority(object):
    """
    Define levels of priority and comparison between them
    """

    # lowest level, used to define a fallback value
    FALLBACK = 10
    # default level
    NORMAL = 50
    # overriding any previous level
    OVERRIDE = 100


@attr.s(slots=True)
class ConfigValue(object):
    """
    Represent config value in config
    """

    value = attr.ib()  # type: Any

    # only value is important, everything else should be cmd=False
    # source = attr.ib(eq=False)  # type: Optional[str]
    # require_parse = attr.ib(default=True, repr=False, eq=False)  # type: bool
    #
    # deprecate_message = attr.ib(default=None, repr=False, eq=False)
    # warnings = attr.ib(factory=list, repr=False, eq=False)

    # Airflow monitor does not support 'eq' in attr.ib()
    source = attr.ib(default="")  # type: Optional[str]
    require_parse = attr.ib(default=True)  # type: bool
    priority = attr.ib(default=ConfigValuePriority.NORMAL)  # type: int

    # mark this config value should be merged
    extend = attr.ib(default=False)  # type: bool
    deprecate_message = attr.ib(default=None)
    warnings = attr.ib(factory=list)


def override(value):
    """Returns configuration with priority override, which is the top configuration priority."""
    return ConfigValue(value=value, source=None, priority=ConfigValuePriority.OVERRIDE)


def default(value):
    """Set default for a config parametr."""
    return ConfigValue(value=value, source=None, priority=ConfigValuePriority.FALLBACK)


def extend(value):
    """
    You can extend any parameter in your environment with the help of extend.

    Lets assume we have our KubernetesConfig with those labels in our cfg.
    ::

        [kubernetes]
        ...
        labels = {"team": "databand_team"}

    And we want to extend those labels for my tasks.
    You can use extend like this::

        @task(task_config={KubernetesConfig.labels: extend({"owner": "Joey"} )})
        ...
    """
    if not (isinstance(value, list) or isinstance(value, dict)):
        raise ValueError(
            "extend is not supported for type {value_type}.\n help: consider using list or dict instead of {value}".format(
                value_type=type(value), value=value
            )
        )

    return ConfigValue(
        value=value, source=None, priority=ConfigValuePriority.NORMAL, extend=True
    )


def fold_config_value(stack, lower):
    # type: (List[ConfigValue], Optional[ConfigValue]) -> List[ConfigValue]
    """
    Collect config values stack by comparing the top of the stack to the current config value.
    If matching priority - it will try to link them if the top of the stack allow, other wise will drop the lower value.
    Otherwise return a stack with the highest priority.
    """
    if lower:
        bottom = first(stack, default=None)
        if bottom is None or bottom.priority < lower.priority:
            # remember highest priority
            return [lower]

        if bottom.extend:
            return [lower] + stack

    return stack



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import typing

import pytest


OnRequestMarker = typing.NamedTuple(
    "OnRequestMarker", [("name", str), ("flag", str), ("description", str)]
)

ON_REQUEST_TEST_TYPES = [
    OnRequestMarker("slow", "--run-slow", "run more than 30 seconds"),
    # clouds
    OnRequestMarker("gcp", "--run-gcp", "require configured gcp"),
    OnRequestMarker("gcp_k8s", "--run-k8s-gcp", "require configured gcp k8s cluster"),
    OnRequestMarker("aws", "--run-aws", "require configured aws"),
    OnRequestMarker("emr", "--run-emr", "require configured emr"),
    OnRequestMarker("livy", "--run-livy", "require configured livy"),
    OnRequestMarker("awsbatch", "--run-awsbatch", "require configured aws batch"),
    OnRequestMarker("azure", "--run-azure", "require configured azure"),
    # tools
    OnRequestMarker("spark", "--run-spark", "require configured local spark"),
    OnRequestMarker("beam", "--run-beam", "require configured local beam"),
    OnRequestMarker("docker", "--run-docker", "require configured local docker"),
]


def pytest_addoption(parser):
    for m in ON_REQUEST_TEST_TYPES:
        parser.addoption(
            m.flag,
            action="store_true",
            default=False,
            help="run tests that %s" % m.description,
        )
    parser.addoption(
        "--run-all",
        action="store_true",
        default=False,
        help="run all tests regardless mark(slow, gcp, aws..)",
    )


def pytest_collection_modifyitems(config, items):
    skips = {
        m.name: pytest.mark.skip(reason="need --run-%s option to run" % m.name)
        for m in ON_REQUEST_TEST_TYPES
    }

    if config.getoption("--run-all"):
        return

    disabled = {m.name: not config.getoption(m.flag) for m in ON_REQUEST_TEST_TYPES}
    for item in items:
        for m in ON_REQUEST_TEST_TYPES:
            if m.name in item.keywords and disabled[m.name]:
                item.add_marker(skips[m.name])



# © Copyright Databand.ai, an IBM Company 2022

import abc
import typing


if typing.TYPE_CHECKING:
    pass


class Marshaller(object):
    type = None
    file_format = None

    # if true on folder read we use load funcition
    support_directory_direct_read = False
    support_multi_target_direct_read = False
    support_directory_direct_write = False

    clears_types_to_str = False

    @abc.abstractmethod
    def target_to_value(self, target, **kwargs):
        pass

    @abc.abstractmethod
    def value_to_target(self, value, target, **kwargs):
        pass

    def support_direct_access(self, target):
        return target.fs.support_direct_access



# © Copyright Databand.ai, an IBM Company 2022

import abc
import errno
import logging
import os
import warnings

import six

from targets.errors import FileAlreadyExists
from targets.utils.atomic import AtomicLocalFile


logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class FileSystem(object):
    """
    FileSystem abstraction used in conjunction with :py:class:`FileTarget`.

    Typically, a FileSystem is associated with instances of a :py:class:`FileTarget`. The
    instances of the py:class:`FileTarget` will delegate methods such as
    :py:meth:`FileTarget.exists` and :py:meth:`FileTarget.remove` to the FileSystem.

    Methods of FileSystem raise :py:class:`FileSystemException` if there is a problem completing the
    operation.
    """

    name = None
    support_direct_access = False
    _exist_after_write_consistent = True
    local = False

    @classmethod
    def exist_after_write_consistent(cls):
        return cls._exist_after_write_consistent

    @abc.abstractmethod
    def exists(self, path):
        """
        Return ``True`` if file or directory at ``path`` exist, ``False`` otherwise

        :param str path: a path within the FileSystem to check for existence.
        """

    @abc.abstractmethod
    def remove(self, path, recursive=True, skip_trash=True):
        """Remove file or directory at location ``path``

        :param str path: a path within the FileSystem to remove.
        :param bool recursive: if the path is a directory, recursively remove the directory and all
                               of its descendants. Defaults to ``True``.
        """

    def mkdir(self, path, parents=True, raise_if_exists=False):
        """
        Create directory at location ``path``

        Creates the directory at ``path`` and implicitly create parent
        directories if they do not already exist.

        :param str path: a path within the FileSystem to create as a directory.
        :param bool parents: Create parent directories when necessary. When
                             parents=False and the parent directory doesn't
                             exist, raise databand.target.MissingParentDirectory
        :param bool raise_if_exists: raise databand.target.FileAlreadyExists if
                                     the folder already exists.
        """
        raise NotImplementedError(
            "mkdir() not implemented on {0}".format(self.__class__.__name__)
        )

    def mkdir_parent(self, path):
        pass

    def isdir(self, path):
        """
        Return ``True`` if the location at ``path`` is a directory. If not, return ``False``.

        :param str path: a path within the FileSystem to check as a directory.

        *Note*: This method is optional, not all FileSystem subclasses implements it.
        """
        raise NotImplementedError(
            "isdir() not implemented on {0}".format(self.__class__.__name__)
        )

    def listdir(self, path):
        """Return a list of files rooted in path.

        This returns an iterable of the files rooted at ``path``. This is intended to be a
        recursive listing.

        :param str path: a path within the FileSystem to list.

        *Note*: This method is optional, not all FileSystem subclasses implements it.
        """
        raise NotImplementedError(
            "listdir() not implemented on {0}".format(self.__class__.__name__)
        )

    def move(self, path, dest):
        """
        Move a file, as one would expect.
        """
        raise NotImplementedError(
            "move() not implemented on {0}".format(self.__class__.__name__)
        )

    def rename_dont_move(self, path, dest):
        """
        Potentially rename ``path`` to ``dest``, but don't move it into the
        ``dest`` folder (if it is a folder).  This relates to :ref:`AtomicWrites`.

        This method has a reasonable but not bullet proof default
        implementation.  It will just do ``move()`` if the file doesn't
        ``exists()`` already.
        """
        warnings.warn(
            "File system {} client doesn't support atomic mv.".format(
                self.__class__.__name__
            )
        )
        if self.exists(dest):
            raise FileAlreadyExists()
        self.move(path, dest)

    def rename(self, *args, **kwargs):
        """
        Alias for ``move()``
        """
        self.move(*args, **kwargs)

    def copy(self, path, dest, **kwargs):
        """
        Copy a file or a directory with contents.
        Currently, LocalFileSystem and MockFileSystem support only single file
        copying but S3Client copies either a file or a directory as required.
        """
        raise NotImplementedError(
            "copy() not implemented on {0}".format(self.__class__.__name__)
        )

    def move_from_local(self, local_path, dest, **kwargs):
        self.copy_from_local(local_path, dest, **kwargs)
        os.remove(local_path)

    def copy_from_local(self, local_path, dest, **kwargs):
        if os.path.isdir(local_path):
            for path, subdirs, files in os.walk(local_path):
                for file in files:
                    # construct the full local path
                    local_file_path = os.path.join(local_path, path, file)
                    relative_path = os.path.relpath(local_file_path, local_path)
                    remote_path = os.path.join(dest, relative_path)
                    self.copy_from_local_file(local_file_path, remote_path, **kwargs)
        else:
            self.copy_from_local_file(local_path, dest, **kwargs)

    def copy_from_local_file(self, local_path, dest, **kwargs):
        raise NotImplementedError(
            "copy_from_local_file() not implemented on {0}".format(
                self.__class__.__name__
            )
        )

    def download(self, path, location, **kwargs):
        logger.info("download %s to %s", path, location)
        parent_location = os.path.dirname(location)

        # parent_location can be empty string if `location` has no dirname, e.g -
        # "some_file_name.txt" or "some_whole_dir"
        if parent_location and not os.path.exists(parent_location):
            try:
                os.makedirs(parent_location)
            except OSError as err:
                # somebody already created the path
                if err.errno != errno.EEXIST:
                    raise

        if self.isdir(path):
            if not os.path.exists(location):
                os.mkdir(location)
            for f in self.listdir(path):
                relative_path = os.path.relpath(f, path)
                local_path = os.path.join(location, relative_path)
                # prevent reference to the folder itself (.)
                if f != path:
                    self.download(f, local_path, **kwargs)
        else:
            self.download_file(path, location, **kwargs)

    def download_file(self, path, location, **kwargs):
        raise NotImplementedError(
            "download_file() not implemented on {0}".format(self.__class__.__name__)
        )

    def open_read(self, path, mode="r"):
        raise NotImplementedError(
            "open_read() not implemented on {0}".format(self.__class__.__name__)
        )

    def open_write(self, path, mode="w", **kwargs):
        return AtomicLocalFile(path, self, mode=mode, **kwargs)

    def move_dir(self, path, dest):
        raise NotImplementedError(
            "move_dir() not implemented on {0}".format(self.__class__.__name__)
        )

    def make_tmp(self):
        raise NotImplementedError(
            "make_tmp() not implemented on {0}".format(self.__class__.__name__)
        )

    def make_tmp_dir(self):
        raise NotImplementedError(
            "make_tmp_dir() not implemented on {0}".format(self.__class__.__name__)
        )



# © Copyright Databand.ai, an IBM Company 2022

from typing import TYPE_CHECKING, Any, Dict

import six

from dbnd._core.constants import ParamValidation
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task.base_task import _BaseTask
from dbnd._core.task_build.task_metaclass import TaskMetaclass


if TYPE_CHECKING:
    from dbnd._core.parameter.parameter_definition import ParameterDefinition
    from dbnd._core.parameter.parameter_value import Parameters, ParameterValue
    from dbnd._core.task_build.task_definition import TaskDefinition


@six.add_metaclass(TaskMetaclass)
class _TaskWithParams(_BaseTask):
    _conf__scoped_params = True  # enable/disable ParameterScope.CHILDREN scope params
    _conf_auto_read_params = True  # enables auto-read value of data params.

    def __init__(
        self,
        task_name,
        task_definition,
        task_signature_obj,
        task_params,  # type: Parameters
        task_config_layer,
        task_config_override,
        task_enabled=True,
        task_sections=None,
        task_call_source=None,
        task_children_scope_params=None,  # type: Dict[str, ParameterValue]
    ):
        super(_TaskWithParams, self).__init__(
            task_name=task_name,
            task_definition=task_definition,
            task_signature_obj=task_signature_obj,
            task_params=task_params,
        )

        self.task_enabled = task_enabled  # relevant only for orchestration

        # configuration data
        self.task_sections = task_sections
        self.task_config_override = task_config_override
        self.task_config_layer = task_config_layer

        self.task_children_scope_params = task_children_scope_params

        self.task_call_source = task_call_source

        for param_value in self.task_params.get_param_values():
            param_value.task = self
            object.__setattr__(self, param_value.name, param_value.value)

    task_definition = None  # type: TaskDefinition

    # user can override this with his configuration
    defaults = None  # type: Dict[ParameterDefinition, Any]

    validate_no_extra_params = parameter.enum(ParamValidation).system(
        default=ParamValidation.disabled,
        description="validate that all configured keys for a task have a matching parameter definition",
    )

    def __setattr__(self, name, value):
        p = (
            self.task_params.get_param_value(name)
            if hasattr(self, "task_params")
            else None
        )
        if p:
            p._update_param_value_from_task_set(value)

        object.__setattr__(self, name, value)

    def _update_property_on_parameter_value_set(self, name, value):
        """called from param.update_value() , we don't want to call back p._update_param_value_from_task_set()"""
        object.__setattr__(self, name, value)

    def _initialize(self):
        pass

    def _validate(self):
        """
        will be called after after object is created
        :return:
        """
        return

    def clone(self, cls=None, output_params_to_clone=None, **kwargs):
        """
        Creates a new instance from an existing instance where some of the args have changed.

        There's at least two scenarios where this is useful (see test/clone_test.py):

        * remove a lot of boiler plate when you have recursive dependencies and lots of args
        * there's task inheritance and some logic is on the base class

        :param cls:
        :param kwargs:
        :return:
        """
        if cls is None:
            cls = self.__class__
        output_params_to_clone = output_params_to_clone or []
        new_k = {}

        # copy only fields that exists in target class
        for param_name, param_class in six.iteritems(
            cls.task_definition.task_param_defs
        ):
            if param_class.is_output() and param_name not in output_params_to_clone:
                continue
            if hasattr(self, param_name):
                new_k[param_name] = getattr(self, param_name)
        new_k["task_name"] = self.task_name

        new_k.update(kwargs)
        return cls(**new_k)

    @classmethod
    def get_task_family(cls):
        return cls.task_definition.task_family

    @classmethod
    def get_full_task_family(cls):
        return cls.task_definition.full_task_family



# © Copyright Databand.ai, an IBM Company 2022

import logging
import signal
import time

from contextlib import contextmanager


logger = logging.getLogger(__name__)


class TimeoutError(Exception):
    pass


@contextmanager
def timeout(seconds, handler=None):
    def timeout_handler(signum, frame):
        raise TimeoutError()

    handler = handler or timeout_handler
    original_handler = signal.signal(signal.SIGALRM, timeout_handler)

    try:
        original_alarm = signal.alarm(seconds)
        if 0 < original_alarm < seconds:  # existing alarm is going to fire earlier
            signal.alarm(original_alarm)
        start_time = time.time()
        yield
    finally:
        if original_alarm:
            original_alarm -= round(time.time() - start_time)
        signal.alarm(original_alarm)
        signal.signal(signal.SIGALRM, original_handler)


def wait_until(predicate, wait_timeout, period=1):
    try:
        with timeout(wait_timeout, handler=lambda *args: None):
            start_time = time.time()
            while not predicate():
                time.sleep(period)
                logger.debug(round(time.time() - start_time))
            return round(time.time() - start_time) or 1
    except TimeoutError:
        pass
    return False



# © Copyright Databand.ai, an IBM Company 2022

from datetime import timedelta

from dbnd._core.utils.timezone import utcnow
from dbnd._vendor import pendulum


def airflow_datetime_str(datetime):
    return pendulum.instance(datetime).to_iso8601_string(extended=True)


_current_unique_execution_date_salt = 0


def unique_execution_date():
    """
    create unique exec date for every run ( so it runs! )
    it should not success datetime.now()
    """
    global _current_unique_execution_date_salt
    _current_unique_execution_date_salt += 1
    if _current_unique_execution_date_salt == 1:
        return utcnow()  # just return current utc

    # task can't run ahead of time! deadlock in scheduler so we go back 1 sec
    # and add unique salt - every time we have new execution date
    return (
        utcnow()
        - timedelta(seconds=1)
        + timedelta(microseconds=_current_unique_execution_date_salt)
    )


def parse_datetime(value, default=None):
    return pendulum.parse(value) if value else default



# © Copyright Databand.ai, an IBM Company 2022

from typing import Optional


def build_conn_path(conn_type, hostname=None, port=None, path=None):
    # type: (str, Optional[str], Optional[int], Optional[str]) -> str
    return "{type}://{hostname}{port}/{path}".format(
        type=conn_type,
        hostname=hostname if hostname else "",
        port=":" + str(port) if port else "",
        path=path.strip("/") if path else "",
    )



# © Copyright Databand.ai, an IBM Company 2022

import uuid

from contextlib import contextmanager


TRACING_ID = uuid.uuid4()


@contextmanager
def new_tracing_id():
    global TRACING_ID
    prev_identifier = TRACING_ID
    TRACING_ID = uuid.uuid4()
    try:
        yield TRACING_ID
    finally:
        TRACING_ID = prev_identifier


def get_tracing_id():
    return TRACING_ID



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import enum

import attr
import six

from dbnd._core.utils.timezone import make_aware, utcnow


RESULT_PARAM = "result"

CURRENT_DATETIME = utcnow()
CURRENT_TIME_STR = CURRENT_DATETIME.strftime("%Y%m%d_%H%M%S")
CURRENT_DATE = CURRENT_DATETIME.date()


class EnvLabel(object):  # env label
    dev = "dev"
    test = "test"
    staging = "stage"
    qa = "qa"
    prod = "prod"


class CloudType(object):
    local = "local"
    gcp = "gcp"
    aws = "aws"
    azure = "azure"


class TaskExecutorType(object):
    local = "local"


class OutputMode(object):
    regular = "regular"
    prod_immutable = "prod_immutable"


class _ConfigParamContainer(object):
    _type_config = True

    @classmethod
    def is_type_config(self, cls):
        # we can't use issubclass as there are "generic" types that will fail this check
        return getattr(cls, "_type_config", False)


class _TaskParamContainer(object):
    pass


class DescribeFormat(object):
    short = "short"
    long = "long"
    verbose = "verbose"


# Compute Types


class EnumWithAll(enum.Enum):
    @classmethod
    def all(cls):
        return list(cls)

    @classmethod
    def all_values(cls):
        return [x.value for x in cls]

    @classmethod
    def all_names(cls):
        return [x.name for x in cls]


class SparkClusters(EnumWithAll):
    local = "local"
    dataproc = "dataproc"
    databricks = "databricks"
    emr = "emr"
    qubole = "qubole"


class ApacheBeamClusterType(object):
    local = "local"
    dataflow = "dataflow"


class ClusterPolicy(object):
    NONE = "none"
    CREATE = "create"
    KILL = "kill"
    EPHERMAL = "ephermal"

    ALL = [NONE, CREATE, KILL, EPHERMAL]


class EmrClient(object):
    LIVY = "livy"
    STEP = "step"


class TaskType(object):
    pipeline = "pipeline"
    python = "python"
    spark = "spark"
    pyspark = "pyspark"
    dataflow = "dataflow"
    docker = "docker"


class TaskRunState(EnumWithAll):
    SCHEDULED = "scheduled"
    QUEUED = "queued"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"

    CANCELLED = "cancelled"
    SHUTDOWN = "shutdown"

    DEFERRED = "deferred"

    UPSTREAM_FAILED = "upstream_failed"
    SKIPPED = "skipped"
    UP_FOR_RETRY = "up_for_retry"

    REMOVED = "removed"

    @staticmethod
    def final_states():
        return TaskRunState.finished_states() | {
            TaskRunState.UPSTREAM_FAILED,
            TaskRunState.SKIPPED,
            TaskRunState.REMOVED,
        }

    @staticmethod
    def final_states_str():
        return [s.value for s in TaskRunState.final_states()]

    @staticmethod
    def finished_states():
        return {TaskRunState.SUCCESS, TaskRunState.FAILED, TaskRunState.CANCELLED}

    @staticmethod
    def finished_states_str():
        return [s.value for s in TaskRunState.finished_states()]

    @staticmethod
    def direct_fail_states():
        return {TaskRunState.FAILED, TaskRunState.CANCELLED}

    @staticmethod
    def fail_states():
        return {
            TaskRunState.FAILED,
            TaskRunState.CANCELLED,
            TaskRunState.UPSTREAM_FAILED,
        }

    @staticmethod
    def states_lower_case():
        return [state.name.lower() for state in TaskRunState]

    @staticmethod
    def parse(state):
        return TaskRunState(state) if state else None

    def is_after(self, old_state):
        # type: (TaskRunState) -> bool
        # Verify, in case of a race condition that we select the most recent state
        ORDER_MAP = {
            TaskRunState.SCHEDULED: 1,
            TaskRunState.QUEUED: 2,
            TaskRunState.RUNNING: 3,
        }
        new_state_order = ORDER_MAP.get(self, 4)
        old_state_order = ORDER_MAP.get(old_state, 4)

        if new_state_order > old_state_order:
            return True

        # A change in the UI, mark as success or mark as failed
        return self in (TaskRunState.FAILED, TaskRunState.SUCCESS) and self != old_state


REUSED = "reused"


class RunState(EnumWithAll):
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"
    SHUTDOWN = "shutdown"
    CANCELLED = "cancelled"
    SCHEDULED = "scheduled"

    @staticmethod
    def finished_states_str():
        return [
            TaskRunState.SUCCESS.value,
            TaskRunState.FAILED.value,
            TaskRunState.CANCELLED.value,
        ]


class AlertStatus(EnumWithAll):
    TRIGGERED = "TRIGGERED"
    RESOLVED = "RESOLVED"
    ACKNOWLEDGED = "ACKNOWLEDGED"

    @classmethod
    def not_resolved(cls):
        return [cls.TRIGGERED.value, cls.ACKNOWLEDGED.value]


class AlertErrorPolicy(object):
    none = ""
    all = "all"
    task_only = "task_only"


class SystemTaskName(object):
    driver_submit = "dbnd_driver_submit"
    driver = "dbnd_driver"
    task_submit = "dbnd_task_submit"

    driver_and_submitter = {driver_submit, driver}


@attr.s
class _DbndDataClass(object):
    def asdict(self, filter=None):
        return attr.asdict(self, recurse=False, filter=filter)


HEARTBEAT_DISABLED = make_aware(datetime.datetime.fromtimestamp(0))


class ParamValidation(EnumWithAll):
    warn = "warn"
    error = "error"
    disabled = "disabled"


class DbndDatasetOperationType(EnumWithAll):
    read = "read"
    write = "write"
    delete = "delete"


class DbndTargetOperationType(EnumWithAll):
    init = "init"
    read = "read"
    write = "write"
    reuse = "reuse"
    log = "log"
    log_hist = "log_hist"
    delete = "delete"


class DbndTargetOperationStatus(EnumWithAll):
    OK = "OK"
    NOK = "NOK"


class SystemMetrics(EnumWithAll):
    Duration = "Duration"
    TotalDuration = "Total Duration"
    ComputeTime = "Compute Time"
    State = "State"

    @staticmethod
    def duration_metrics():
        """Used to select metrics for removal during metrics re-generation"""
        return [
            s.value
            for s in [
                SystemMetrics.Duration,
                SystemMetrics.TotalDuration,
                SystemMetrics.ComputeTime,
            ]
        ]

    @staticmethod
    def get_description(metric_name):
        if metric_name == SystemMetrics.Duration.value:
            return "Execution time of the latest run attempt of this task"
        elif metric_name == SystemMetrics.TotalDuration.value:
            return "Sum the duration of all task run attempts of this task"
        elif metric_name == SystemMetrics.ComputeTime.value:
            return "Sum of the duration of all task attempts related to this run"
        else:
            return None


class UpdateSource(EnumWithAll):
    # `dbnd` source is only for orchestration
    dbnd = "dbnd"
    airflow_monitor = "airflow_monitor"
    airflow_tracking = "airflow_tracking"
    azkaban_tracking = "azkaban_tracking"
    generic_tracking = "generic_tracking"

    def __eq__(self, other):
        if isinstance(other, UpdateSource):
            return self.value == other.value
        elif isinstance(other, six.string_types):
            return str(self) == other or str(self.value) == other

        return False

    @classmethod
    def is_tracking(cls, source):
        return source in [UpdateSource.airflow_tracking, UpdateSource.azkaban_tracking]


class MetricSource(object):
    user = "user"
    system = "system"
    histograms = "histograms"
    spark = "spark"
    dbt = "dbt"

    @classmethod
    def all(cls):
        return [cls.user, cls.system, cls.histograms, cls.spark]

    @classmethod
    def default_sources(cls):
        return [cls.user, cls.system, cls.histograms, cls.spark]

    @classmethod
    def default_sources_str(cls):
        return ",".join(cls.default_sources())


AD_HOC_DAG_PREFIX = "DBND_RUN."


class AlertSeverity(object):
    CRITICAL = "CRITICAL"
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"

    @classmethod
    def values(cls):
        return [cls.CRITICAL, cls.HIGH, cls.MEDIUM, cls.LOW]


class AlertDefOperator(EnumWithAll):
    ANOMALY = "ML"
    RANGE = "range"
    EQUAL = "=="
    NOT_EQUAL = "!="
    GREATER_THAN = ">"
    NOT_LESS_THAN = ">="
    LESS_THAN = "<"
    NOT_GREATER_THAN = "<="


TASK_ESSENCE_ATTR = "task_essence"


class TaskEssence(enum.Enum):
    ORCHESTRATION = "orchestration"
    TRACKING = "tracking"
    CONFIG = "config"

    @classmethod
    def is_task_cls(self, cls):
        return (
            hasattr(cls, TASK_ESSENCE_ATTR)
            and getattr(cls, TASK_ESSENCE_ATTR) != self.CONFIG
        )

    def is_instance(self, obj):
        """
        Checks if the object is include in the essence group.
        >>> TaskEssence.TRACKING.is_instance(obj)
        """
        return (
            hasattr(obj, TASK_ESSENCE_ATTR) and getattr(obj, TASK_ESSENCE_ATTR) == self
        )


class AirflowEnvironment(EnumWithAll):
    ON_PREM = "on_prem"



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.plugin.dbnd_plugins import pm
from dbnd._core.utils.basics.load_python_module import _load_module
from dbnd._core.utils.seven import fix_sys_path_str


logger = logging.getLogger(__name__)

_dbnd_plugins_registered = False


def register_dbnd_plugins():
    if get_dbnd_project_config().is_no_plugins:
        return

    global _dbnd_plugins_registered
    if _dbnd_plugins_registered:
        return
    _dbnd_plugins_registered = True

    fix_sys_path_str()
    if not get_dbnd_project_config().disable_pluggy_entrypoint_loading:
        pm.load_setuptools_entrypoints("dbnd")
        pm.check_pending()


def register_dbnd_user_plugins(user_plugin_modules):
    for plugin_module in user_plugin_modules:
        module = _load_module(plugin_module, "plugin:%s" % plugin_module)
        pm.register(module)

        base_msg = "Plugin %s" % plugin_module
        if getattr(module, "__version__", None):
            base_msg += " v%s" % module.__version__

        logger.info(base_msg + " loaded...")



# © Copyright Databand.ai, an IBM Company 2022

import typing

from typing import Any, Dict, Iterable, List, Optional, Tuple

import attr

from dbnd._core.errors import DatabandError
from dbnd._core.parameter.constants import ParameterScope


if typing.TYPE_CHECKING:
    from dbnd._core.parameter.parameter_definition import ParameterDefinition
    from dbnd._core.task.task_with_params import _TaskWithParams


@attr.s(slots=True)
class ParameterValue(object):
    """Holds the runtime value of a parameter and information about its calculation"""

    # reference to the parameter definition
    parameter = attr.ib()  # type: ParameterDefinition
    # the calculated value
    value = attr.ib()  # type: Any

    # name of all the information sources which used to calculate the value
    source = attr.ib()  # type: str
    source_value = attr.ib()  # type: Any

    task = attr.ib(default=None)  # type: _TaskWithParams
    parsed = attr.ib(default=True)  # type: bool

    # any warnings caught while calculating the value
    warnings = attr.ib(factory=list)

    @property
    def name(self):
        return self.parameter.name

    def update_param_value(self, value):
        self.value = value
        if self.task:
            self.task._update_property_on_parameter_value_set(self.name, value)

    def _update_param_value_from_task_set(self, value):
        # handler for task.param=value code
        # we should not call task here, otherwise, we get recursion
        # we assume that value is already set in task body
        self.value = value


def fold_parameter_value(left, right):
    # type: (ParameterValue, Optional[ParameterValue]) -> ParameterValue
    """Merge the values of the two parameters and combine any information needed to identify this merge"""
    if not right or right.value is None:
        return left

    if not left.parameter.value_type.support_merge:
        raise ValueError(
            "value type {} not supporting merge".format(left.parameter.value_type)
        )

    if left.parameter.value_type != right.parameter.value_type:
        raise ValueError(
            "can't merge two value with different types left={} right={}".format(
                left.parameter.value_type, right.parameter.value_type
            )
        )

    new_value = left.parameter.value_type.merge_values(left.value, right.value)

    return ParameterValue(
        parameter=left.parameter,
        value=new_value,
        source=",".join((left.source, right.source)),
        source_value=left.source_value,
        warnings=left.warnings + right.warnings,
    )


@attr.s(slots=True)
class ParameterFilter(object):
    """
    Represent filter settings for Parameters
    """

    significant_only = attr.ib(default=False)
    input_only = attr.ib(default=False)
    output_only = attr.ib(default=False)
    user_only = attr.ib(default=False)

    param_cls = attr.ib(default=None)
    scope = attr.ib(default=None)  # type: Optional[ParameterScope]


class ParameterFilters(object):
    CHILDREN = ParameterFilter(scope=ParameterScope.children)
    USER = ParameterFilter(user_only=True)
    OUTPUTS = ParameterFilter(output_only=True)
    INPUTS = ParameterFilter(input_only=True)
    USER_INPUTS = ParameterFilter(user_only=True, input_only=True)
    USER_OUTPUTS = ParameterFilter(user_only=True, output_only=True)
    SIGNIFICANT_ONLY = ParameterFilter(significant_only=True)
    SIGNIFICANT_INPUTS = ParameterFilter(input_only=True, significant_only=True)
    NONE = ParameterFilter()


class Parameters(object):
    def __init__(self, source, param_values):
        # type: (str,  List[ParameterValue]) -> None

        self.source = source
        self._param_values = param_values
        self._param_values_map = {p.parameter.name: p for p in self._param_values}

    def get_param_value_safe(self, param_name):  # type: (str) -> ParameterValue
        pv = self.get_param_value(param_name)
        if not pv:
            raise DatabandError(
                "Param value '%s' is not found at %s" % (param_name, self.source)
            )
        return pv

    def update_param_value(self, param_name, value):
        self.get_param_value_safe(param_name).update_param_value(value)

    def get_param_value(self, param_name):  # type: (str) -> Optional[ParameterValue]
        return self._param_values_map.get(param_name, None)

    def get_param(self, param_name):  # type: (str) -> Optional[ParameterDefinition]
        p = self.get_param_value(param_name)
        if p:
            return p.parameter
        return None

    def get_value(self, param_name):
        p = self.get_param_value(param_name)
        if not p:
            raise DatabandError(
                "Parameter '%s' not found at %s" % (param_name, self.source)
            )
        return p.value

    def _filter_params(self, param_filter):
        # type: (ParameterFilter)-> Iterable[ParameterValue]
        result = self._param_values  # type: Iterable[ParameterValue]

        pf = param_filter or ParameterFilters.NONE
        if pf.param_cls is not None:
            result = filter(lambda p: isinstance(p.parameter, pf.param_cls), result)
        if pf.significant_only:
            result = filter(lambda p: p.parameter.significant, result)
        if pf.input_only:
            result = filter(lambda p: not p.parameter.is_output(), result)
        if pf.output_only:
            result = filter(lambda p: p.parameter.is_output(), result)
        if pf.user_only:
            result = filter(lambda p: not p.parameter.system, result)
        if pf.scope:
            result = filter(lambda p: p.parameter.scope == pf.scope, result)
        return result

    def get_params(self, param_filter=None):
        # type: (Optional[ParameterFilter])-> List[ParameterDefinition]
        return [pv.parameter for pv in self._filter_params(param_filter=param_filter)]

    def get_param_values(self, param_filter=None):
        # type: (Optional[ParameterFilter])-> List[ParameterValue]
        return list(self._filter_params(param_filter=param_filter))

    def get_params_with_value(self, param_filter=None):
        # type: (Optional[ParameterFilter])-> List[ Tuple[ParameterDefinition, Any]]
        return [
            (pv.parameter, pv.value)
            for pv in self._filter_params(param_filter=param_filter)
        ]

    # TODO: change name to "to_string"
    def get_params_serialized(self, param_filter=None):
        # type: (Optional[ParameterFilter])-> List[ Tuple[ParameterDefinition, Any]]
        return [
            (pv.parameter.name, pv.parameter.signature(pv.value))
            for pv in self._filter_params(param_filter=param_filter)
        ]

    def get_params_signatures(self, param_filter=None):
        # type: (Optional[ParameterFilter])-> List[ Tuple[ParameterDefinition, Any]]
        return [
            (pv.parameter.name, pv.parameter.signature(pv.value))
            for pv in self._filter_params(param_filter=param_filter)
        ]

    def to_env_map(self, task, *param_names):
        # type: (_TaskWithParams, List[str]) -> Dict[str, str]
        if param_names:
            params = [self.get_param_value(param_name) for param_name in param_names]
        else:
            params = self.get_param_values()
        return {
            p.parameter.get_env_key(task.task_name): p.parameter.to_str(p.value)
            for p in params
        }

    def get_param_env_key(self, task, param_name):
        return self.get_param(param_name).get_env_key(task.task_name)

    def __iter__(self):
        # type: (Parameters) -> Iterable[ParameterValue]
        """helper function to iterate all the params with it's definition, value and meta"""
        for pv in self._param_values:
            yield pv

    def as_key_value_dict(self):
        return {p.name: p.value for p in self.get_param_values()}



# © Copyright Databand.ai, an IBM Company 2022

from contextlib import contextmanager

import six


@contextmanager
def patched_properties(obj, properties):
    """
    This method is a context-manager which can be called as below:

    .. code-block: python
                with no_unpicklable_properties(self,("some_property", patch_value)):
                    do_something(self)

    """
    patched_properties = {}
    new_properties = []
    for name, value in properties:
        if hasattr(obj, name):
            patched_properties[name] = getattr(obj, name)
        else:
            new_properties.append(name)

        setattr(obj, name, value)

    yield obj

    for name, value in six.iteritems(patched_properties):
        setattr(obj, name, value)

    for name in new_properties:
        delattr(obj, name)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._vendor.termcolor import colored


# values should be calculated at run time - so colored function could be removed if needed
def error_separator():
    return colored("--------------------------------------", color="red")


def error_separator_small():
    return colored("- - -", on_color="on_red")


def bold(value):
    return colored(value, attrs=["bold"])


def underline(value):
    return colored(value, attrs=["underline"])


def red(value):
    return colored(value, color="red")


def cyan(value):
    return colored(value, color="cyan")



# © Copyright Databand.ai, an IBM Company 2022

from collections import namedtuple

from more_itertools import first

from dbnd._core.current import get_databand_context
from dbnd.api.query_params import build_query_api_params, create_filters_builder
from dbnd.api.shared_schemas.scheduled_job_schema import ScheduledJobSchemaV2


ScheduledJobNamedTuple = namedtuple(
    "ScheduledJobNamedTuple", ScheduledJobSchemaV2().fields.keys()
)
ScheduledJobNamedTuple.__new__.__defaults__ = (None,) * len(
    ScheduledJobNamedTuple._fields
)


def post_scheduled_job(scheduled_job_dict, ctx=None):
    ctx = ctx or get_databand_context()
    schema = ScheduledJobSchemaV2(strict=False)
    data, _ = schema.dump({"DbndScheduledJob": scheduled_job_dict})
    response = ctx.databand_api_client.api_request(
        "/api/v1/scheduled_jobs", data, method="POST", no_prefix=True
    )
    return schema.load(data=response).data


def patch_scheduled_job(scheduled_job_dict, ctx=None):
    ctx = ctx or get_databand_context()
    schema = ScheduledJobSchemaV2(strict=False)
    data, _ = schema.dump({"DbndScheduledJob": scheduled_job_dict})
    response = ctx.databand_api_client.api_request(
        "/api/v1/scheduled_jobs?name=%s" % scheduled_job_dict["name"],
        data,
        method="PATCH",
        no_prefix=True,
    )
    return schema.load(data=response).data


def delete_scheduled_job(scheduled_job_name, revert=False, ctx=None):
    ctx = ctx or get_databand_context()
    ctx.databand_api_client.api_request(
        "/api/v1/scheduled_jobs?name=%s&revert=%s"
        % (scheduled_job_name, str(revert).lower()),
        None,
        method="DELETE",
        no_prefix=True,
    )


def get_scheduled_jobs(
    name_pattern=None, from_file_only=None, include_deleted=False, ctx=None
):
    ctx = ctx or get_databand_context()
    query_params = build_query_api_params(
        filters=build_scheduled_job_filter(
            name_pattern=name_pattern,
            from_file_flag=from_file_only,
            include_deleted_flag=include_deleted,
        )
    )
    return _get_scheduled_jobs(ctx.databand_api_client, query_params)


def set_scheduled_job_active(name, value, ctx=None):
    ctx = ctx or get_databand_context()
    ctx.databand_api_client.api_request(
        "/api/v1/scheduled_jobs/set_active?name=%s&value=%s"
        % (name, str(value).lower()),
        None,
        method="PUT",
        no_prefix=True,
    )


def _get_scheduled_jobs(client, query_api_args):
    endpoint = "?".join(["scheduled_jobs", query_api_args])
    schema = ScheduledJobSchemaV2(strict=False)
    response = client.api_request(endpoint=endpoint, data=None, method="GET")
    return schema.load(data=response["data"], many=True).data


def get_scheduled_job_by_name(job_name):
    query_params = build_query_api_params(
        filters=build_scheduled_job_filter(job_name=job_name)
    )

    job_result = _get_scheduled_jobs(
        get_databand_context().databand_api_client, query_params
    )
    try:
        return first(job_result)["DbndScheduledJob"]
    except ValueError:
        return None


def is_scheduled_job_exists(job_name):
    return bool(get_scheduled_job_by_name(job_name))


build_scheduled_job_filter = create_filters_builder(
    job_name=("name", "eq"),
    name_pattern=("name", "like"),
    from_file_flag=("from_file", "eq"),
    include_deleted_flag=("deleted_from_file", "eq"),
)



# © Copyright Databand.ai, an IBM Company 2022

import datetime

import six

from dbnd._core.errors import ParameterError
from dbnd._core.utils.date_interval import DateInterval
from targets.values import ValueType


class DateIntervalValueType(ValueType):
    """
    A ValueType whose value is a :py:class:`~databand.date_interval.DateInterval`.

    Date Intervals are specified using the ISO 8601 date notation for dates
    (eg. "2015-11-04"), months (eg. "2015-05"), years (eg. "2015"), or weeks
    (eg. "2015-W35"). In addition, it also supports arbitrary date intervals
    provided as two dates separated with a dash (eg. "2015-11-04-2015-12-04").
    """

    type = DateInterval

    def parse_from_str(self, s):
        """
        Parses a :py:class:`~databand.date_interval.DateInterval` from the input.

        see :py:mod:`databand.date_interval`
          for details on the parsing of DateIntervals.
        """
        # TODO: can we use xml.utils.iso8601 or something similar?

        from dbnd._core.utils import date_interval as d

        for cls in [d.Year, d.Month, d.Week, d.Date, d.Custom]:
            i = cls.parse(s)
            if i:
                return i

        raise ValueError("Invalid date interval - could not be parsed")


class TimeDeltaValueType(ValueType):
    """
    Class that maps to timedelta using strings in any of the following forms:

     * ``n {w[eek[s]]|d[ay[s]]|h[our[s]]|m[inute[s]|s[second[s]]}`` (e.g. "1 week 2 days" or "1 h")
        Note: multiple arguments must be supplied in longest to shortest unit order
     * ISO 8601 duration ``PnDTnHnMnS`` (each field optional, years and months not supported)
     * ISO 8601 duration ``PnW``

    See https://en.wikipedia.org/wiki/ISO_8601#Durations
    """

    type = datetime.timedelta

    def parse_from_str(self, input):
        """
        Parses a time delta from the input.

        See :py:class:`TimeDeltaValueType` for details on supported formats.
        """
        result = _parseIso8601(input)
        if not result:
            result = _parseSimple(input)
        if result is not None:
            return result
        else:
            raise ParameterError("Invalid time delta - could not parse %s" % input)

    def to_str(self, x):
        """
        Converts datetime.timedelta to a string

        :param x: the value to serialize.
        """
        weeks = x.days // 7
        days = x.days % 7
        hours = x.seconds // 3600
        minutes = (x.seconds % 3600) // 60
        seconds = (x.seconds % 3600) % 60
        result = " ".join(
            [
                "%s%s" % (value, n)
                for n, value in [
                    ("w", weeks),
                    ("d", days),
                    ("h", hours),
                    ("m", minutes),
                    ("s", seconds),
                ]
                if value
            ]
        )
        if not result:
            result = "0d"
        # result = "{} w {} d {} h {} m {} s".format(weeks, days, hours, minutes, seconds)
        return result


def _apply_regex(regex, input):
    import re

    re_match = re.match(regex, input)
    if re_match and any(re_match.groups()):
        kwargs = {}
        has_val = False
        for k, v in six.iteritems(re_match.groupdict(default="0")):
            val = int(v)
            if val > -1:
                has_val = True
                kwargs[k] = val
        if has_val:
            return datetime.timedelta(**kwargs)


def _parseIso8601(input):
    def field(key):
        return r"(?P<%s>\d+)%s" % (key, key[0].upper())

    def optional_field(key):
        return "(%s)?" % field(key)

    # A little loose: ISO 8601 does not allow weeks in combination with other fields, but this regex does (as does python timedelta)
    regex = "P(%s|%s(T%s)?)" % (
        field("weeks"),
        optional_field("days"),
        "".join([optional_field(key) for key in ["hours", "minutes", "seconds"]]),
    )
    return _apply_regex(regex, input)


def _parseSimple(input):
    keys = ["weeks", "days", "hours", "minutes", "seconds"]
    # Give the digits a regex group name from the keys, then look for text with the first letter of the key,
    # optionally followed by the rest of the word, with final char (the "s") optional
    regex = "".join(
        [r"((?P<%s>\d+) ?%s(%s)?(%s)? ?)?" % (k, k[0], k[1:-1], k[-1]) for k in keys]
    )
    return _apply_regex(regex, input)



# © Copyright Databand.ai, an IBM Company 2022

import os
import tempfile


# see https://blog.thedataincubator.com/2016/06/testing-jupyter-notebooks/
def notebook_run(path):
    """Execute a notebook via nbconvert and collect output.
    :returns (parsed nb object, execution errors)
    """
    import nbformat

    from nbconvert import nbconvertapp

    # file can't be open multiple times on windows
    # https://docs.python.org/2/library/tempfile.html
    with tempfile.NamedTemporaryFile(suffix=".ipynb", delete=False) as fout:
        tmp_file_name = fout.name
    dirname, _ = os.path.split(path)

    nbconvertapp.main(
        [
            "nbconvert",
            "--to",
            "notebook",
            "--execute",
            "--ExecutePreprocessor.timeout=60",
            "--output",
            tmp_file_name,
            path,
        ]
    )

    nb = nbformat.read(tmp_file_name, nbformat.current_nbformat)

    if os.path.exists(tmp_file_name):
        os.remove(tmp_file_name)

    errors = [
        output
        for cell in nb.cells
        if "outputs" in cell
        for output in cell["outputs"]
        if output.output_type == "error"
    ]

    return nb, errors



# © Copyright Databand.ai, an IBM Company 2022

import json
import logging

from dbnd._core.errors.base import DatabandApiError
from dbnd._core.utils.basics.text_banner import TextBanner, safe_tabulate
from dbnd._vendor import click
from dbnd.api.airflow_sync import (
    archive_airflow_instance,
    create_airflow_instance,
    edit_airflow_instance,
    list_synced_airflow_instances,
    unarchive_airflow_instance,
)


BYTES_IN_KB = 1024
logger = logging.getLogger(__name__)


@click.group()
@click.pass_context
def airflow_sync(ctx):
    """Manage synced Airflow instances"""


@airflow_sync.command("list")
def list_airflow_instances():
    try:
        instances = list_synced_airflow_instances()
    except (LookupError, DatabandApiError) as e:
        logger.warning(e)
    else:
        print_table("Synced Airflow instances", instances)


def print_table(header, instances_list):
    banner = TextBanner(header)
    banner.write(build_instances_table(instances_list))
    logger.info(banner.get_banner_str())


def build_instances_table(instances_data):
    extract_keys = (
        "name",
        "is_sync_enabled",
        "tracking_source_uid",
        "base_url",
        "external_url",
        "api_mode",
    )
    headers = (
        "Name",
        "Active",
        "Tracking source UID",
        "Base url",
        "External url",
        "Api Mode",
        "Instance Type",
    )
    table_data = []
    for server_info in instances_data:
        instance_row = [getattr(server_info, key, "") for key in extract_keys]
        instance_row.append(
            "Airflow" if getattr(server_info, "fetcher") == "web" else "GC Composer"
        )
        table_data.append(instance_row)
    return safe_tabulate(table_data, headers)


@airflow_sync.command()
@click.option(
    "--url", "-u", help="base url for instance", type=click.STRING, required=True
)
@click.option(
    "--external-url",
    "-e",
    help="External url for instance",
    type=click.STRING,
    default=None,
)
@click.option(
    "--fetcher",
    "-f",
    help="Fetcher to use for data (web, db)",
    type=click.Choice(["web", "db"], case_sensitive=False),
    default="db",
)
@click.option("--env", help="Environment", type=click.STRING, default=None)
@click.option(
    "--include-sources",
    help="Monitor source code for tasks",
    type=click.BOOL,
    is_flag=True,
)
@click.option(
    "--dag-ids",
    help="List of specific dag ids (separated with comma) that monitor will fetch only from them",
    type=click.STRING,
    default=None,
)
@click.option(
    "--last-seen-dag-run-id",
    help="Id of the last dag run seen in the Airflow database",
    type=click.STRING,
    default=None,
)
@click.option(
    "--last-seen-log-id",
    help="Id of the last log seen in the Airflow database",
    type=click.STRING,
    default=None,
)
@click.option("--name", help="Name for the syncer", type=click.STRING, default=None)
@click.option(
    "--generate-token",
    help="Generate access token for the syncer, value is token lifespan (in seconds)",
    type=click.INT,
    default=None,
)
@click.option(
    "--config-file-output",
    help="Store syncer config json to file",
    type=click.File("w"),
    default="-",
)
@click.option(
    "--with-auto-alerts",
    help="Create syncer with auto alerts config",
    type=click.BOOL,
    is_flag=True,
)
@click.option(
    "--include-logs-bytes-from-head",
    help="Include the number of bytes from the head of the log file",
    type=click.IntRange(min=0, max=8096),
    default=0,
)
@click.option(
    "--include-logs-bytes-from-end",
    help="Include the number of bytes from the end of the log file",
    type=click.IntRange(min=0, max=8096),
    default=0,
)
@click.option(
    "--dag-run-bulk-size",
    help="DAG run bulk size for the syncer",
    type=click.INT,
    default=None,
)
@click.option(
    "--start-time-window",
    help="Start time window for the syncer (in days)",
    type=click.INT,
    default=None,
)
def add(
    url,
    external_url,
    fetcher,
    env,
    include_sources,
    dag_ids,
    last_seen_dag_run_id,
    last_seen_log_id,
    name,
    generate_token,
    config_file_output,
    with_auto_alerts,
    include_logs_bytes_from_head,
    include_logs_bytes_from_end,
    dag_run_bulk_size,
    start_time_window,
):
    try:
        if not env:
            env = name

        system_alert_definitions = {
            "failed_state": bool(with_auto_alerts),
            "ml_run_duration": bool(with_auto_alerts),
            "run_schema_change": bool(with_auto_alerts),
        }

        monitor_config = {
            "include_sources": bool(include_sources),
            "log_bytes_from_end": include_logs_bytes_from_end * BYTES_IN_KB,
            "log_bytes_from_head": include_logs_bytes_from_head * BYTES_IN_KB,
        }

        if dag_run_bulk_size:
            monitor_config["dag_run_bulk_size"] = dag_run_bulk_size

        if start_time_window:
            monitor_config["start_time_window"] = start_time_window

        config_json = create_airflow_instance(
            url,
            external_url,
            fetcher,
            env,
            dag_ids,
            last_seen_dag_run_id,
            last_seen_log_id,
            name,
            generate_token,
            system_alert_definitions,
            monitor_config,
        )
        if config_file_output:
            json.dump(config_json, config_file_output, indent=4)
    except DatabandApiError as e:
        logger.warning("failed with - {}".format(e.response))
    else:
        logger.info("Successfully added %s", url)


@airflow_sync.command()
@click.option(
    "--tracking-source-uid",
    "-u",
    help='Tracking source uid of the edited airflow syncer. (you can get this with the "list" command)',
    type=click.STRING,
    required=True,
)
@click.option(
    "--url", "-u", help="base url for instance", type=click.STRING, required=True
)
@click.option(
    "--external-url",
    "-e",
    help="External url for instance",
    type=click.STRING,
    default=None,
)
@click.option(
    "--fetcher",
    "-f",
    help="Fetcher to use for data (web, db)",
    type=click.Choice(["web", "db"], case_sensitive=False),
    default="db",
)
@click.option("--env", help="Environment", type=click.STRING, default=None)
@click.option(
    "--include-sources",
    help="Don't monitor source code for tasks",
    type=click.BOOL,
    is_flag=True,
)
@click.option(
    "--dag-ids",
    help="List of specific dag ids (separated with comma) that monitor will fetch only from them",
    type=click.STRING,
    default=None,
)
@click.option(
    "--last-seen-dag-run-id",
    help="Id of the last dag run seen in the Airflow database",
    type=click.STRING,
    default=None,
)
@click.option(
    "--last-seen-log-id",
    help="Id of the last log seen in the Airflow database",
    type=click.STRING,
    default=None,
)
@click.option("--name", help="Name for the syncer", type=click.STRING, default=None)
@click.option(
    "--with-auto-alerts",
    help="Create syncer with auto alerts config",
    type=click.BOOL,
    is_flag=True,
)
@click.option(
    "--include-logs-bytes-from-head",
    help="Include the number of bytes from the head of the log file",
    type=click.IntRange(min=0, max=8096),
    default=0,
)
@click.option(
    "--include-logs-bytes-from-end",
    help="Include the number of bytes from the end of the log file",
    type=click.IntRange(min=0, max=8096),
    default=0,
)
@click.option(
    "--dag-run-bulk-size",
    help="DAG run bulk size for the syncer",
    type=click.INT,
    default=None,
)
@click.option(
    "--start-time-window",
    help="Start time window for the syncer (in days)",
    type=click.INT,
    default=None,
)
def edit(
    tracking_source_uid,
    url,
    external_url,
    fetcher,
    env,
    include_sources,
    dag_ids,
    last_seen_dag_run_id,
    last_seen_log_id,
    name,
    with_auto_alerts,
    include_logs_bytes_from_head,
    include_logs_bytes_from_end,
    dag_run_bulk_size,
    start_time_window,
):
    try:
        system_alert_definitions = {
            "failed_state": bool(with_auto_alerts),
            "ml_run_duration": bool(with_auto_alerts),
            "run_schema_change": bool(with_auto_alerts),
        }

        monitor_config = {
            "include_sources": bool(include_sources),
            "log_bytes_from_end": include_logs_bytes_from_end * BYTES_IN_KB,
            "log_bytes_from_head": include_logs_bytes_from_head * BYTES_IN_KB,
        }

        if dag_run_bulk_size:
            monitor_config["dag_run_bulk_size"] = dag_run_bulk_size

        if start_time_window:
            monitor_config["start_time_window"] = start_time_window

        edit_airflow_instance(
            tracking_source_uid,
            url,
            external_url,
            fetcher,
            env,
            dag_ids,
            last_seen_dag_run_id,
            last_seen_log_id,
            name,
            system_alert_definitions,
            monitor_config,
        )
    except DatabandApiError as e:
        logger.warning("failed with - {}".format(e.response))
    else:
        logger.info("Successfully edited %s", url)


@airflow_sync.command()
@click.option(
    "--url",
    "-u",
    help="base url of an instance to archive",
    type=click.STRING,
    required=True,
)
def archive(url):
    try:
        archive_airflow_instance(url)
    except DatabandApiError as e:
        logger.warning("failed with - {}".format(e.response))
    else:
        logger.info("Archived instance %s", url)


@airflow_sync.command()
@click.option(
    "--url",
    "-u",
    help="base url of an instance to unarchive",
    type=click.STRING,
    required=True,
)
def unarchive(url):
    try:
        unarchive_airflow_instance(url)
    except DatabandApiError as e:
        logger.warning("failed with - {}".format(e.response))
    else:
        logger.info("Unarchived instance %s", url)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from collections import Counter, defaultdict
from functools import partial
from typing import Any, Optional, Union

import six

from dbnd._core.errors import DatabandSystemError, friendly_error
from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled
from dbnd._core.task.task import Task
from dbnd._core.utils.traversing import traverse
from targets.base_target import Target
from targets.multi_target import MultiTarget
from targets.target_factory import target
from targets.types import Path


logger = logging.getLogger(__name__)


def _try_get_task_from_airflow_op(value):
    if is_airflow_enabled():
        from dbnd_airflow.dbnd_task_executor.converters import try_operator_to_dbnd_task

        return try_operator_to_dbnd_task(value)


def _to_task(value):
    if isinstance(value, Task):
        return value

    if isinstance(value, Target):
        if value.source_task:
            return value.source_task
        if isinstance(value, MultiTarget):
            tasks = [_to_task(t) for t in value.targets]
            return [t for t in tasks if t is not None]

        # we are not going to create "task" for the target,
        # if it doesn't have an source
        return None

    airflow_op_task = _try_get_task_from_airflow_op(value)
    if airflow_op_task:
        return airflow_op_task

    logger.debug("Can't convert '%s' to task.", value)
    return None


def to_tasks(obj_or_struct):
    return traverse(obj_or_struct, convert_f=_to_task, filter_none=True)


def _to_target(value, from_string_kwargs=None):
    if value is None:
        return value

    if isinstance(value, Target):
        return value

    if isinstance(value, Task):
        if value.task_definition.single_result_output:
            return value.result
        return value.task_outputs

    airflow_op_task = _try_get_task_from_airflow_op(value)
    if airflow_op_task:
        return airflow_op_task.task_outputs

    if isinstance(value, six.string_types):
        from_string_kwargs = from_string_kwargs or {}
        return target(value, **from_string_kwargs)

    if isinstance(value, Path):
        from_string_kwargs = from_string_kwargs or {}
        return target(str(value), **from_string_kwargs)

    raise friendly_error.failed_to_convert_value_to_target(value)


def to_targets(obj_or_struct, from_string_kwargs=None):
    return traverse(
        obj_or_struct,
        convert_f=partial(_to_target, from_string_kwargs=from_string_kwargs),
        filter_none=True,
    )


def targets_to_str(obj_or_struct):
    return traverse(obj_or_struct, convert_f=target_to_str, filter_none=True)


def target_to_str(obj):
    # type: (Any) -> str
    if isinstance(obj, six.string_types):
        return obj
    return str(obj)


def tasks_summary(tasks):
    if not tasks:
        return ""
    tasks_count = [
        "%s %s" % (count, key)
        for key, count in Counter((t.task_name for t in tasks)).most_common()
    ]
    return "{stats}  - total {total}".format(
        total=len(tasks), stats=",".join(tasks_count)
    )


def tasks_to_ids_set(tasks):
    return set(t.task_id for t in tasks)


def calculate_friendly_task_ids(tasks):
    tasks_by_name = defaultdict(list)
    for t in tasks:
        tasks_by_name[t.task_name].append(t)
    task_friendly_ids = {}
    for tasks in tasks_by_name.values():
        sorted_tasks = sorted(tasks, key=lambda t: t.task_creation_id)
        for i, task in enumerate(sorted_tasks):
            if i > 0 or len(sorted_tasks) > 1:
                task_af_id = "{}_{}".format(task.task_name, i)
            else:
                task_af_id = task.task_name
            task_friendly_ids[task.task_id] = task_af_id
    return task_friendly_ids


def get_task_name_safe(task_or_task_name):
    # type: (Union[Task, str]) -> str
    if task_or_task_name is None or isinstance(task_or_task_name, six.string_types):
        return task_or_task_name

    if isinstance(task_or_task_name, Task):
        return task_or_task_name.task_name
    raise DatabandSystemError(
        "Can't calculate task name from %s - %s",
        task_or_task_name,
        type(task_or_task_name),
    )


def get_project_name_safe(project_name, task_or_task_name):
    # type: (Union[Task, str], str) -> Optional[str]
    if project_name:
        return project_name

    if task_or_task_name is None or isinstance(task_or_task_name, six.string_types):
        return None

    if isinstance(task_or_task_name, Task) and hasattr(task_or_task_name, "project"):
        return task_or_task_name.project
    return None
